"""Classes used for dispersing shapes inside a :class:`~vcams.voxelpart.VoxelPart`."""

import logging
import time
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Iterable, Callable

import matplotlib
import matplotlib.pyplot as plt
import numpy as np  # This is intentional to simplify the formulae.
from numpy import random, max, abs, isscalar, sum, inf, ndarray
from scipy.stats import truncnorm

from .shape import ShapeArray
from ..logger_conf import setup_dispersion_logger, LogWithoutFormatContext

matplotlib.use('TkAgg')  # From https://stackoverflow.com/a/73788178/7180705
logger = logging.getLogger(__name__)


class TooManyDispersionAttemptsError(Exception):
    pass


class TooManyDispersionTrialsError(Exception):
    pass


class TooManyDispersionGenerationsError(Exception):
    pass


class TooMuchDeviationError(Exception):
    pass


class TooManyValueGenerationAttemptsError(Exception):
    pass


class SuitableNumShapesNotFoundError(Exception):
    pass


class BaseListDispersion(ABC):
    """Abstract base class for dispersions that contain a list of values.
    Subclasses are used for defining various dispersions.
    """

    @property
    def actual_mean(self):
        """Actual mean of the values in the instance."""
        return self.values.mean()

    @property
    def actual_std(self):
        """Actual standard deviation of the values in the instance."""
        return self.values.std()

    @property
    def actual_variance(self):
        """Actual variance of the values in the instance."""
        return self.values.var()

    @property
    def _repr_float_length(self):
        """The optimal number of floating point decimal places.
        Used for representing the instance in text format."""
        if len(self) == 0:
            return 0
        else:
            repr_float_length = len(str(int(max(abs((self.actual_mean, self.actual_std,
                                                     self.actual_variance)))))) + 1
        return repr_float_length + 4  # Add four decimal places.

    def __iter__(self):
        return self.values.__iter__()

    def __getitem__(self, index):
        return self.values[index]

    def __len__(self):
        return len(self.values)

    def __mul__(self, other):
        return ManualListDispersion(self.values * other)

    def __init__(self):
        """Constructor for the :class:`BaseListDispersion` abstract base class.
        It should be called by subclasses."""
        self.values = np.array([])
        """A numpy array containing the values in the instance."""

    @staticmethod
    def _set_plot_legend():
        """Set the legend for the plots created by the :meth:`BaseListDispersion.plot` method."""
        plt.gca().legend(frameon=False, prop={'family': 'monospace'},  # bbox_to_anchor=(1.04, 1),
                         fontsize=2, numpoints=1, loc='upper right')
        # plt.tight_layout()

    def plot(self, num_bins: int, plot_actual_normal_curve: bool = False):
        """Plot a histogram of the dispersion object.

        Args:
            num_bins: Number of bins for the histogram of actual values.
            plot_actual_normal_curve: Whether the actual normal distribution of the values should be plotted.
                                      Defaults to False.
        """
        if self.__len__() < 1:
            raise ValueError('The object is empty. Has it been initialized?')
        plt.plot([], [], ' ', label=f'$\\bf{{Dispersion Type: {type(self).__name__}}}$')
        plt.hist(self.values, num_bins, density=True, label='Actual Values Histogram')
        plt.xlabel('Value')
        plt.ylabel('Frequency')

        if plot_actual_normal_curve:
            x = np.linspace(-3 * self.actual_std + self.actual_mean, 3 * self.actual_std + self.actual_mean, 100)
            y = (np.exp(-np.power((x - self.actual_mean) / self.actual_std, 2.0) / 2)
                 / (np.sqrt(2.0 * np.pi) * self.actual_std))
            plt.plot(x, y, label='Normal Distribution (Actual)')

        self._set_plot_legend()
        plt.show()


class ManualListDispersion(BaseListDispersion):
    """A Dispersion list defined manually using a list.
    This is simply a list and is only created to have a uniform naming.
    """

    def __init__(self, values: Iterable | ndarray):
        """Constructor for the :class:`ManualListDispersion` class.
        It takes an iterable, creates a *NumPy* array from it, and then flattens it.
        """
        super().__init__()
        self.values = np.array(values).flatten()

    def __repr__(self):
        return (f'{self.__class__.__name__} Instance:\n'
                f'    Number of Values: {len(self.values)}\n'
                f'    Actual Mean:      {self.actual_mean:{self._repr_float_length}.4f}\n'
                f'    Actual STD:       {self.actual_std:{self._repr_float_length}.4f}\n'
                f'    Actual Variance:  {self.actual_variance:{self._repr_float_length}.4f}')

    # noinspection PyMethodOverriding
    def plot(self, num_bins: int):
        """Plot a histogram of the dispersion object.

        Args:
            num_bins: Number of bins for the histogram of actual values.
        """
        BaseListDispersion.plot(self, num_bins, plot_actual_normal_curve=True)
        plt.plot([], [], ' ', label=self.__repr__().split('\n', 1)[1])
        BaseListDispersion._set_plot_legend()
        plt.show()


class BaseNormalDistributionDispersion(BaseListDispersion):
    """Abstract base class for dispersion classes that generate a list of values
    with a `normal (Gaussian) distribution <https://en.wikipedia.org/wiki/Normal_distribution>`_.

    Instances are defined using a target mean and standard deviation.
    Afterwards, a list is created using a random number generator.
    Instances are iterable and iterating over them is the same
    as iterating over their *values* attribute.

    If the number of values in an instance is specified at creation,
    the list of values will be populated. Otherwise, generation of the list will be deferred.
    At any point, the instances :meth:`generate_values` method can be called to
    create any number of values which will replace existing values.

    This is an abstract base class and cannot be instantiated.
    Child classes define their own :meth:`generate_values` method which determines
    the details of how the list of values is generated.
    """

    def __init__(self, target_mean: float, target_std: float,
                 num_values: int = None, max_absolute_pct_error: int = 10):
        """Constructor for the :class:`BaseNormalDistributionDispersion` class.

        Args:
            target_mean: Target mean for the randomly generated values.
                         Actual mean will be stored in *actual_mean*.
            target_std: Target standard deviation for the randomly generated values.
                        Actual standard deviation will be stored in *actual_mean*.
            num_values: Number of values to be generated and stored in the *values* property.
                        Defaults to *None* which defers value generation and
                        requires a call to :meth:`generate_values` to generate the values.
            max_absolute_pct_error: See :meth:`BaseNormalDistributionDispersion._qc_dispersion`.
                                    Defaults to 10%.
        """
        super().__init__()

        self.target_mean = target_mean
        """See :meth:`__init__`'s arguments."""
        self.target_std = target_std
        """See :meth:`__init__`'s arguments."""
        self.max_absolute_pct_error = max_absolute_pct_error
        """See :meth:`__init__`'s arguments."""

        if num_values is not None:
            self.generate_values(num_values)
        else:
            self.values = []

    @property
    def _repr_float_length(self):
        """The optimal number of floating point decimal places.
        Used for representing the instance in text format."""
        # This is overridden from BaseListDispersion.
        if len(self) == 0:
            repr_float_length = len(str(int(max(abs((self.target_mean, self.target_std)))))) + 1
        else:
            repr_float_length = len(str(int(max(abs((self.target_mean, self.actual_mean,
                                                     self.target_std, self.actual_std,
                                                     self.actual_variance)))))) + 1
        return repr_float_length + 4  # Add four decimal places.

    def __repr__(self):
        if len(self) == 0:
            return (f'{self.__class__.__name__} Instance:\n'
                    f'    The instance is empty. Use generate_values(num_values) to populate it.\n'
                    f'    Target Mean:      {self.target_mean:{self._repr_float_length}.4f}\n'
                    f'    Target STD:       {self.target_std:{self._repr_float_length}.4f}')
        else:
            mean_pct_error = 100 * (self.actual_mean - self.target_mean) / self.target_mean
            std_pct_error = 100 * (self.actual_std - self.target_std) / self.target_std
            return (f'{self.__class__.__name__} Instance:\n'
                    f'    Number of Values: {len(self.values)}\n'
                    f'    Target Mean:      {self.target_mean:{self._repr_float_length}.4f}\n'
                    f'    Actual Mean:      {self.actual_mean:{self._repr_float_length}.4f} ({mean_pct_error:+.4f}%)\n'
                    f'    Target STD:       {self.target_std:{self._repr_float_length}.4f}\n'
                    f'    Actual STD:       {self.actual_std:{self._repr_float_length}.4f} ({std_pct_error:+.4f}%)\n'
                    f'    Actual Variance:  {self.actual_variance:{self._repr_float_length}.4f}')

    # noinspection PyMethodOverriding
    def plot(self, num_bins: int, plot_actual_normal_curve: bool = False):
        """Plot a histogram of the dispersion object.

        Args:
            num_bins: Number of bins for the histogram of actual values.
            plot_actual_normal_curve: Whether the actual normal distribution of the values should be plotted.
                                      Defaults to False.
        """
        BaseListDispersion.plot(self, num_bins, plot_actual_normal_curve)
        x = np.linspace(-3 * self.target_std + self.target_mean, 3 * self.target_std + self.target_mean, 1000)
        y = (np.exp(-np.power((x - self.target_mean) / self.target_std, 2.0) / 2)
             / (np.sqrt(2.0 * np.pi) * self.target_std))
        plt.plot(x, y, label='Normal Distribution (Target)')
        plt.plot([], [], ' ', label=self.__repr__().split('\n', 1)[1])
        BaseListDispersion._set_plot_legend()
        plt.show()

    @abstractmethod
    def _generate_values_once(self, num_values: int, qc_results=True):
        """Abstract method for generating a single set of random values
        based on the instance's attributes.
        This method can be called as many times as necessary to regenerate the instance's
        *values* list and change its *num_values*.

        Args:
            num_values: Number of values to be generated and stored in the *values* property.
            qc_results: If *True*, the values are quality controlled
                        using :meth:`_qc_dispersion`. Defaults to *True*.
        """
        pass

    def generate_values(self, num_values: int, qc_results=True, max_attempts=10000):
        """Generate values based on the normal dispersion's parameters.

        Args:
            num_values: Number of values to be generated and stored in the *values* property.
            qc_results: If *True*, the values are quality controlled
                        using :meth:`BaseNormalDistributionDispersion._qc_dispersion`.
                        Defaults to *True*.
            max_attempts: The maximum number of attempts for generation of valid values.
                          If exceeded, :class:`TooManyValueGenerationAttemptsError` is raised.
                          Note that some parameters require much more tries
                          because the distribution keeps getting rejected
                          by the quality control process.

        Raises:
            TooManyValueGenerationAttemptsError: Too many attempts were made for value generation.
        """
        for i in range(max_attempts):
            try:
                self._generate_values_once(num_values=num_values, qc_results=qc_results)
                return
            except TooMuchDeviationError:
                continue
        raise TooManyValueGenerationAttemptsError(f'Too many attempts ({max_attempts}) made '
                                                  f'for generating valid values.')

    def _qc_dispersion(self, min_size: int = 0, max_absolute_pct_error: int | None = None):
        """Control the quality of the dispersion using two tests:

        - The number of values in the dispersion must be more than *min_size*. Defaults to 0.
        - The absolute percent error (APE) between target and actual values of mean and standard deviation,
          defined by :math:`|\\frac{x_{actual} - x_{target}}{x_{target}}| \\times 100\\%`,
          must be equal or less than *max_absolute_pct_error*.
          Defaults to *None* which uses the instance's *max_absolute_pct_error*.

          If the instance does not pass QC, :class:`TooMuchDeviationError` is raised.
        """
        if max_absolute_pct_error is None:
            max_absolute_pct_error = self.max_absolute_pct_error

        if len(self) < min_size:
            raise ValueError(f'The dispersion instance does not pass QC '
                             f'because it is too few values ({len(self)}<{min_size}).')
        std_ape = abs((self.actual_std - self.target_std) / self.target_std) * 100
        if std_ape > max_absolute_pct_error:
            raise TooMuchDeviationError(f'The absolute percent error between actual '
                                        f'and target standard deviation is {std_ape:.2f}% '
                                        f'but should be lower than {max_absolute_pct_error}%. '
                                        f'There are {len(self)} values.')
        mean_ape = abs((self.actual_mean - self.target_mean) / self.target_mean) * 100
        if mean_ape > max_absolute_pct_error:
            raise TooMuchDeviationError(f'The absolute percent error between actual '
                                        f'and target mean is {mean_ape:.2f}% '
                                        f'but should be lower than {max_absolute_pct_error}%. '
                                        f'There are {len(self)} values.')


class NormalDistributionDispersion(BaseNormalDistributionDispersion):
    """A dispersion class that generates a list of values with
    a `normal (Gaussian) distribution <https://en.wikipedia.org/wiki/Normal_distribution>`_.

    The generated distribution will be truly normal,
    but may include values that are negative or too small.
    This will cause issues for some purposes such as geometrical features.
    The :class:`TruncatedNormalDistributionDispersion` class
    can fix this problem but its values will not be truly normal.

    This is a subclass of :class:`BaseNormalDistributionDispersion`.
    See its docs for other details.
    """

    def _generate_values_once(self, num_values: int, qc_results=True):
        """Generate a single set of random values with the given mean and standard deviation.
        This method can be called as many times as necessary to regenerate the instance's
        *values* list and change its *num_values*.

        This function uses NumPy's `random.default_rng().normal` function.

        Args:
            num_values: Number of values to be generated and stored in the *values* property.
            qc_results: If *True*, the values are quality controlled
                        using :meth:`~BaseNormalDistributionDispersion._qc_dispersion`.
                        Defaults to *True*.
        """
        if num_values < 1:
            raise ValueError('num_values must be bigger than 0.')
        self.values = random.default_rng().normal(loc=self.target_mean, scale=self.target_std, size=num_values)
        if qc_results:
            self._qc_dispersion()


class TruncatedNormalDistributionDispersion(BaseNormalDistributionDispersion):
    """A dispersion class that generates a list of values with
    a `truncated normal (Gaussian) distribution <https://en.wikipedia.org/wiki/Truncated_normal_distribution>`_.

    A regular normal distribution created by the :class:`NormalDistributionDispersion` class
    has the problem that it may include values that are negative or too small
    which will cause issues for some purposes such as geometrical features.
    This class allows for defining boundaries for the start and end of the distribution
    which solves this problem.
    However, there are two problems:

      1. The results will not be truly normal.
      2. The generated values may not pass the quality control process
         which means that many attempts may be necessary
         when calling the :meth:`~BaseNormalDistributionDispersion.generate_values` method.

    This is a subclass of :class:`BaseNormalDistributionDispersion`.
    See its docs for other details.
    """

    def __init__(self, target_mean: float, target_std: float,
                 bound_a: float = -inf, bound_b: float = inf, num_values: int = None):
        """Constructor for the :class:`TruncatedNormalDistributionDispersion` class.

        Args:
            target_mean: Target mean for the randomly generated values.
                         Actual mean will be stored in *actual_mean*.
            target_std: Target standard deviation for the randomly generated values.
                        Actual standard deviation will be stored in *actual_mean*.
            bound_a:    The beginning of the range from which values should be drawn.
            bound_b:    The end of the range from which values should be drawn.
            num_values: Number of values to be generated and stored in the *values* property.
                        Defaults to *None* which defers value generation and
                        requires a call to :meth:`~BaseNormalDistributionDispersion.generate_values`
                        to generate the values.
        """
        # target_mean and target_std are assigned in super().__init__.
        self.bound_a = bound_a
        """See :meth:`__init__`'s arguments."""
        self.bound_b = bound_b
        """See :meth:`__init__`'s arguments."""
        # Boundaries a and b are around the mean. They must be moved to the real scale.
        self._truncnorm_a = (bound_a - target_mean) / target_std
        self._truncnorm_b = (bound_b - target_mean) / target_std
        super().__init__(target_mean, target_std, num_values)

    def plot(self, num_bins: int, plot_equivalent_normal_curve: bool = True):
        """Plot a histogram of the dispersion object.

        Args:
            num_bins: Number of bins for the histogram of actual values.
            plot_equivalent_normal_curve: Whether an equivalent (untruncated) normal distribution
                                          should be plotted based on the parameters.
                                          Defaults to True.
        """
        BaseListDispersion.plot(self, num_bins)

        x = np.linspace(-3 * self.target_std + self.target_mean, 3 * self.target_std + self.target_mean, 1000)
        y = truncnorm.pdf(x, a=self._truncnorm_a, b=self._truncnorm_b,
                          loc=self.target_mean, scale=self.target_std)
        plt.plot(x, y, label='Truncated Normal Distribution (Target)')
        if plot_equivalent_normal_curve:
            y_eq_normal = (np.exp(-np.power((x - self.target_mean) / self.target_std, 2.0) / 2)
                           / (np.sqrt(2.0 * np.pi) * self.target_std))
            plt.plot(x, y_eq_normal, label='Equivalent Normal Distribution (Target)')

        plt.plot([], [], ' ', label=self.__repr__().split('\n', 1)[1])
        BaseListDispersion._set_plot_legend()
        plt.show()

    def _generate_values_once(self, num_values: int, qc_results=True):
        """Generate a single set of random values based on a truncated normal distribution
        with the given mean and standard deviation.
        This method can be called as many times as necessary to regenerate the instance's
        *values* list and change its *num_values*.

        This function uses SciPy's `stats.truncnorm.rvs` function.

        Args:
            num_values: Number of values to be generated and stored in the *values* property.
            qc_results: If *True*, the values are quality controlled
                        using the :meth:`~BaseNormalDistributionDispersion._qc_dispersion` method.
                        Defaults to *True*.
        """
        # See __init__ for _truncnorm_a and _truncnorm_b.
        self.values = truncnorm.rvs(a=self._truncnorm_a, b=self._truncnorm_b,
                                    loc=self.target_mean, scale=self.target_std,
                                    size=num_values)
        if qc_results:
            self._qc_dispersion()


class RandomDispersion:
    """A dispersion class that generates a single random float value in a half-open interval.

    The constructor (:meth:`__init__`) takes three arguments,
    which are *low*, *high*, and boundary, and then calculates
    :attr:`actual_low` and :attr:`actual_high` using the following equations:

    .. math::

       \\begin{cases}
       \\text{actual_low} &=& \\text{low} &+& \\text{boundary}\\\\
       \\text{actual_high} &=& \\text{high} &-& \\text{boundary}
       \\end{cases}

    This class uses *numpy.random.uniform()* in the half-open interval
    :math:`[\\text{actual_low}, \\text{actual_high})`.

    Instances of this class are callables, meaning that they can be called like a function.
    For example:

    .. code-block:: python

       rand_disp_ins = RandomDispersion(low=2, high=10)
       v1 = rand_disp_ins()  # A random value between 2 and 10.
       v2 = rand_disp_ins()  # Another random value between 2 and 10.

    """

    def __init__(self, low: float, high: float, boundary: float = 0):
        """Constructor for the :class:`RandomDispersion` class.

        Args:
            low: Lower boundary of the output interval.
            high: Upper boundary of the output interval.
            boundary: A boundary applied to the *low* and *high* arguments
                      to calculate :attr:`actual_low` and :attr:`actual_high`.
        """
        self.low = low
        self.high = high
        self.boundary = boundary
        self.actual_low = low + boundary
        """Actual lower boundary from which the random number is generated.
        Values will be equal or greater than this number."""
        self.actual_high = high - boundary
        """Actual upper boundary from which the random number is generated.
        Values will be less than this number."""

    def __call__(self):
        return random.uniform(low=self.actual_low, high=self.actual_high, size=None)

    def __repr__(self):
        return (f'{self.__class__.__name__} Instance:\n'
                f'    Low:         {self.low}\n'
                f'    High:        {self.high}\n'
                f'    Boundary:    {self.boundary}\n'
                f'    Actual Low:  {self.actual_low}\n'
                f'    Actual High: {self.actual_high}\n')

    def plot(self):
        """A nonfunctional method that raises a *NotImplementedError*
        because it is not applicable to the :class:`RandomDispersion` class."""
        raise NotImplementedError('This function is not available for the RandomDispersion class '
                                  'because it only return a random scalar.')


class ShapeDispersionArray(ShapeArray):
    """Class for an array of shapes that are dispersed in the workspace.
    For each shape, its xyz coordinates are randomly assigned
    and the rest of the parameters may be predefined or randomly generated.

    This is, in fact, a subclass of :class:`~.shape.ShapeArray`
    which can disperse shapes inside itself.

    The array may contain any number of shapes of any class as long as they
    are subclasses of :class:`.shape.BaseShape` and have the same *dim* attribute.

    When constructing an instance, a :class:`~vcams.voxelpart.VoxelPart` instance
    can be passed which allows for defining a *background* in the shape array.
    This means that the shapes will only be dispersed in the nonempty regions
    of the *VoxelPart* instance.
    """

    def __init__(self, dim: str, part=None,
                 part_shape: tuple[int, int, int] = None,
                 voxel_size: tuple[float, float, float] = None,
                 wrap_mask: bool = True, num_bound_pixels: int = 0,
                 short_msg: bool = True):
        """Constructor for the *ShapeDispersionArray* class.

        Args:
            dim: Dimensionality of the shape array which determines the shapes that
                 can be added to the shape array. Valid values are '2D' and '3D'.
            part (VoxelPart | None): The *VoxelPart* instance based on which the shape array is created.
                                     If *None*, *part_shape* and *voxel_size* must be specified,
                                     otherwise they are ignored. Defaults to *None*.
            part_shape: A tuple containing two or three integers which determine
                        the shape of the returned boolean mask.
                        Defaults to *None* and ignored if *part* is passed.
            voxel_size: A tuple containing two or three floats which determine the size of a voxel
                        in the x, y, and z directions.
                        Defaults to *None* and ignored if *part* is passed.
            wrap_mask: If set to True, the shapes' function is wrapped around
                       the boundaries of the working space, enabling periodic structures.
                       If *wrap_mask* is True, *num_bound_pixels* must be zero.
                       Defaults to True.
            num_bound_pixels: An int specifying the number of pixels to add to the boundary of the base mask.
                              The boundary will become a region that the dispersed shapes cannot touch.
                              If *num_bound_pixels* is non-zero, *wrap_mask* must be False.
                              Defaults to 0.
            short_msg: A boolean specifying whether the placement message should be printed
                       as a single updating line or in many lines with extensive details.
                       Passed to :meth:`._log_placement` Defaults to *True*.
        """

        # Validate the relationship between wrap_mask and num_bound_pixels.
        # The rest is validated when calling super().__init__().
        if wrap_mask and num_bound_pixels != 0:
            raise ValueError('wrap_mask is True and num_bound_pixels is non-zero. This combination in invalid.')

        # Call the parent class's constructor.
        # Note that is_mask_calculation_lazy is set to True.
        super().__init__(dim, part, part_shape, voxel_size,
                         is_mask_calculation_lazy=True, wrap_mask=wrap_mask)

        self.short_msg = short_msg
        """See :meth:`.__init__`."""

        self.shape_requests = []
        """A list of shapes shape classes and related parameters that should be dispersed.
        
        Members of this list are added by the :meth:`add_shape_request` method.
        Each is a list containing the following:
        
          - **cls**: The *class* object for the shape that is requested.
            It is a subclass of :class:`.shape.BaseShape`
            and has the same *dim* attribute (2D or 3D) as the shape array.
            The last two members of the list will be keyword arguments
            that can be used to define the shape. 
          
          - **num_shapes**: Number of requested shapes. This should be an integer.
            If set to *None*, the number can be determined by the object.
            In that case, all members of the list should have *num_shapes* set to *None*.
          
          - **iterable_kwargs**: A dictionary of keyword arguments for *cls*.
            Each dictionary key will be the name of a keyword argument for *cls*,
            and the dictionary value will be the keyword argument's value.
            Here, keys are either iterables or subclasses of :class:`BaseListDispersion`,
            except for :class:`RandomDispersion` which is treated as a scalar.
            Note that the dictionary values should have the same length as *num_shapes*.
          
          - **scalar_kwargs**: A dictionary of keyword arguments for *cls*.
            Each dictionary key will be the name of a keyword argument for *cls*,
            and the dictionary value will be the keyword argument's value.
            Here, keys are either scalars or instances :class:`RandomDispersion`.
            Scalars will be used for all the shapes requested (by this member)
            and instances :class:`RandomDispersion` will be called
            by the appropriate methods to generate random values.
          
        Addition to this list should be done by the :meth:`add_shape_request` method
        and the list is emptied after a successful dispersion.
        Users should not interact with this list. This is not enforced but highly recommended.
        """

        self._dispersion_log_file_path = \
            part._log_file_path.with_stem(part._log_file_path.stem + '_dispersion_log')  # noqa: PyProtectedMember
        self.dispersion_logger = setup_dispersion_logger(part.name, log_file=self._dispersion_log_file_path,
                                                         display_log=True)

        # Add boundary to the base mask so the shapes don't touch the outside.
        if num_bound_pixels:
            self.base_mask[:, :num_bound_pixels] = True
            self.base_mask[:, -num_bound_pixels:] = True
            self.base_mask[:num_bound_pixels, :] = True
            self.base_mask[-num_bound_pixels:, :] = True

    @property
    def num_requested_shapes(self) -> int:
        """Total number of shapes requested in :attr:`shape_requests`.
        This is calculated automatically on the fly.
        If any of the *num_shapes* in *shape_requests* is set to *None*,
        the number -1 is returned."""
        num_shapes_list = [sr[1] for sr in self.shape_requests]
        if None in num_shapes_list:
            return -1
        else:
            return int(sum(num_shapes_list, axis=None))

    def _backup_state(self):
        super()._backup_state()
        self._backup_dict['shape_requests'] = deepcopy(self.shape_requests)

    def _restore_state(self, backup_dict=None):
        super()._restore_state(backup_dict)
        if backup_dict is None:
            backup_dict = self._backup_dict
        self.shape_requests = deepcopy(backup_dict['shape_requests'])

    @staticmethod
    def _test_cls_kwargs(cls, iterable_kwargs, scalar_kwargs):
        """Try to create a shape with a set of iterable_kwargs and scalar_kwargs
        to make sure they are valid. It raises an error if unsuccessful and is otherwise silent.

        This is a method function used by the :meth:`add_shape_request` method
        and should not be directly called by the users."""
        iter0_dict_temp = dict()
        scalar_dict_temp = dict()
        for (k, v) in iterable_kwargs.items():
            if isinstance(v, BaseNormalDistributionDispersion) and (len(v) == 0):
                v.generate_values(1, qc_results=False)  # QC is turned off.
            iter0_dict_temp[k] = v[0]
        for (k, v) in scalar_kwargs.items():
            if isinstance(v, RandomDispersion):
                scalar_dict_temp[k] = v()
            else:
                scalar_dict_temp[k] = v
        try:
            _ = cls(id=-1, **iter0_dict_temp, **scalar_dict_temp)
        except Exception as err:
            exception_msg = (f'The given **kwargs cannot be used with class {cls.__name__}. They are:\n'
                             f'First element of each iterable keyword argument:\n  {iter0_dict_temp}\n'
                             f'Scalar and sample random keyword arguments:\n  {scalar_dict_temp}')
            raise Exception(exception_msg).with_traceback(err.__traceback__)

    def _place_shape_randomly(self, cls, shape_number: int, max_attempts: int,
                              trial_number: int = None, **kwargs):
        """Place a shape in a random position. Placement is tried *max_attempts* number of times
        and if not successful, :class:`TooManyDispersionAttemptsError` is raised.

        This is a private method used by the :meth:`disperse_shapes` method
        and should not be directly called by the users.

        Args:
            cls: The shape class that should be placed. Arguments are passed as *kwargs*.
            shape_number: The number of this shape.
                          This number keeps track of the shapes that are dispersed
                          and is used for logging the progress. it is *not* the shape's eventual ID.
            max_attempts: The maximum number of attempts for placement of a shape.
                          If exceeded, :class:`TooManyDispersionAttemptsError` is raised.
            trial_number: The number for this trial. Used for logging the progress.
            **kwargs:     A dictionary where the keys are the arguments for the shape class
                          and the values are either dispersion objects or scalars.
                          Coordinate arguments should always be passed as :class:`RandomDispersion`
                          instances that are then called to generate a random value for each attempt.
                          This is not enforced.

        Raises:
            TooManyDispersionAttemptsError: Too many attempts were made for this shape.
        """

        # Separate random and scalar kwargs.
        # Note that scalars may have come from an iterable object, but they are scalars now.
        scalar_dict = dict()
        random_object_dict = dict()
        for k, v in kwargs.items():
            if isinstance(v, RandomDispersion):
                random_object_dict[k] = v
            else:  # Assume it's a scalar.
                scalar_dict[k] = v

        # Try to place randomly max_attempts times.
        # Return if successful. Otherwise, raise an error in the end.
        for attempt_number in range(1, max_attempts + 1):
            random_value_dict = dict()
            for k, v in random_object_dict.items():
                random_value_dict[k] = v()  # Note that v (being a RandomDispersion) is *called* here.
            self._log_placement(cls, attempt_number, shape_number, scalar_dict, random_value_dict, trial_number)
            # Try to add a shape to the array. shape_status will be True if successful.
            shape_status = self.add_shape(cls, intersect_ok=False, **scalar_dict, **random_value_dict)
            self._log_shape_placement_status(shape_status)
            if shape_status:
                return
        raise TooManyDispersionAttemptsError(f'Too many dispersion attempts for shape {shape_number}.')

    def _log_placement(self, cls, attempt_number, shape_number, scalar_dict,
                       random_value_dict, trial_number=None):
        """Log the placement of the shape.
        This is a private method used by the :meth:`_place_shape_randomly` method
        and should not be directly called by the users."""
        if trial_number is None:
            trial_str = ''
        else:
            trial_str = f'Trial {trial_number:4d}, '
        shape_num_str = f'Shape {shape_number:{len(str(self.num_requested_shapes))}d}/{self.num_requested_shapes}'
        attempt_num_str = f'Attempt {attempt_number:4d}'
        msg_str = f'{trial_str}{shape_num_str}, {attempt_num_str}'
        if self.short_msg:
            msg_str = msg_str + f', Type {cls.__name__} ... '
        else:
            scalar_items_str = ', '.join(f'{k}={v:2.4f}' for k, v in scalar_dict.items())
            random_items_str = ', '.join(f'{k}={v:2.4f}' for k, v in random_value_dict.items())
            msg_str = msg_str + f', Type {cls.__name__}({scalar_items_str}, {random_items_str}) ... '
        self.dispersion_logger.debug(msg_str)

    def _log_shape_placement_status(self, shape_status):
        """Log the placement status of shape.
        This is a private method used by the :meth:`_place_shape_randomly` method
        and should not be directly called by the users."""
        if shape_status:
            with LogWithoutFormatContext(self.dispersion_logger):
                self.dispersion_logger.debug('Success!\r')
        else:
            with LogWithoutFormatContext(self.dispersion_logger):
                self.dispersion_logger.debug(' Failed!\r')

    def add_shape_request(self, cls, num_shapes: int | None, **kwargs):
        """Add a request for a one or more shapes of a class to be added to the instance.

        Args:
            cls: The shape class that should be placed. Arguments are passed as *kwargs*.
            num_shapes: Number of shapes that are requested with the given arguments.
                        Should either be an integer or *None* so it can be determined using
                        the instance's :meth:`_find_suitable_num_shapes` method.
                        If set to *None*, all other shape requests must also have this argument
                        set to *None*. Defaults to *None*.
            **kwargs: A dictionary where the keys are the arguments for the shape class
                      and the values are either dispersion objects or scalars.
                      The function separates the dictionary into two dictionaries,
                      one containing scalars and the other containing.
                      Note that the length of the iterable dispersion objects
                      must be equal to *num_shapes*.
        """

        # Validate the shape class.
        self._check_shape_class(cls)

        # Split the kwargs dictionary into two groups,
        # that will each be passed to the placement function differently.
        # The keys are the arguments' names.
        # In iterable_kwargs, the value is an iterable whose contents will be passed one by one;
        # and in other_kwargs, the value is a single scalar which will be passed each time.
        iterable_kwargs = dict()
        scalar_kwargs = dict()
        for k, v in kwargs.items():
            if isinstance(v, ManualListDispersion):
                if num_shapes is None:
                    raise ValueError(f'{k} is a ManualListDispersion which is not allowed '
                                     'when num_shapes is None.')
                elif len(v) != num_shapes:
                    raise ValueError(f'{k} is a ManualListDispersion with {len(v)} elements '
                                     f'but num_shapes is {num_shapes}. They should be equal.')
                else:
                    iterable_kwargs[k] = v
            elif isinstance(v, BaseNormalDistributionDispersion):
                if num_shapes is None:  # Values should not be generated for v. Current values are inconsequential.
                    pass
                elif len(v) != 0:  # v is uninitialized and should have its values generated.
                    v.generate_values(num_values=num_shapes)
                else:
                    raise ValueError(f'{k} is an uninitialized subclass of '
                                     f'BaseNormalDistributionDispersion. '
                                     f'Specify num_shapes when defining it. '
                                     f'Did you mean to pass num_shapes=None to the method?')
                iterable_kwargs[k] = v
            elif isscalar(v) or isinstance(v, RandomDispersion):
                scalar_kwargs[k] = v
            else:
                raise ValueError(f'{k} is neither a scalar nor a valid subclass of BaseListDispersion.')

        # Try to create a shape with iterable_kwargs and scalar_kwargs to make sure they're valid.
        ShapeDispersionArray._test_cls_kwargs(cls, iterable_kwargs, scalar_kwargs)

        # Add the shape request as a tuple to the instance's shape_requests list.
        self.shape_requests.append([cls, num_shapes, iterable_kwargs, scalar_kwargs])
        logger.debug(f'Adding a shape request for the {cls.__name__} class.')

    def disperse_shapes(self, max_attempts: int = 5000, max_trials: int = 100, log_main: bool = True):
        """Disperse shapes in the *ShapeDispersionArray* instance according to
        the shape requests. All shape requests are processed
        and then :attr:`shape_requests` is emptied.

        It is necessary for all the members in the instance's :attr:`shape_requests`
        to have valid *num_shapes*, otherwise an error is raised.
        If you need to disperse without knowing the number of shapes
        (based on volume fraction), use the :meth:`disperse_shapes_vf` method instead.

        The function uses the :meth:`_place_shape_randomly` function to place each shape
        in a random position. Placement is tried *max_attempts* number of times
        and if the attempts run out, the *trial* ends and a new trial begins with
        a clean ShapeArray, and the same shapes will be dispersed again.
        If after the given maximum number of trials (*max_trials*) all shapes are not placed,
        :class:`TooManyDispersionTrialsError` will be raised to indicate an error.

        Args:
            max_attempts: The maximum number of attempts for placement of a shape.
                          If exceeded, the trial ends and the process is restarted.
            max_trials:   The maximum number of trials. If exceeded,
                          :class:`TooManyDispersionTrialsError` is raised.
            log_main:     Whether the start and end of dispersion should be logged by the main logger.
                          Defaults to True.

        Raises:
            TooManyDispersionTrialsError: Too many trials were done.
                                          Note that each trials entails many attempts.
        """
        begin_msg = 'Dispersing shapes in the ShapeDispersionArray.'
        self.dispersion_logger.debug(begin_msg + '\n')
        if log_main:
            logger.debug(begin_msg)
        if max_attempts < 1:
            raise ValueError('max_attempts should be >= 1.')
        if max_trials < 1:
            raise ValueError('max_trials should be >= 1.')

        num_shapes_list = [sr[1] for sr in self.shape_requests]
        if None in num_shapes_list:
            raise ValueError(f'Some of the num_shapes in shape_requests are None.'
                             f'The list is: {num_shapes_list}')

        # Backup the state of the ShapeDispersionArray object.
        begin_time = time.perf_counter()
        self._backup_state()
        dispersion_success = False
        for trial_number in range(1, max_trials + 1):
            try:  # Try to run a trial.
                dispersion_success = False
                shape_number = 0
                for sr in self.shape_requests:
                    (cls, num_shapes, iterable_kwargs, scalar_kwargs) = sr
                    # Loop through the list kwargs.
                    for i in range(num_shapes):
                        shape_number += 1
                        # Put the shape's iterable_kwargs into a single dict to be passed as scalars.
                        shape_list_kwargs = dict()
                        for k, v in iterable_kwargs.items():
                            shape_list_kwargs[k] = v[i]
                        # Try to place the shape. Raises TooManyDispersionAttempts if unsuccessful.
                        self._place_shape_randomly(cls, shape_number, max_attempts, trial_number,
                                                   **shape_list_kwargs, **scalar_kwargs)
                # Both loops have run out without error and dispersion is successful.
                dispersion_success = True
                break
            except TooManyDispersionAttemptsError:
                # Too many attempts were made in self._place_shape_randomly().
                # Restore the state of the ShapeDispersionArray object and retry.
                self._restore_state()
        # Either the trial loop has been broken which indicates success,
        # or it has ended which means we have run out of trials without success.
        if dispersion_success:
            elapsed_time = time.perf_counter() - begin_time
            self._log_dispersion_success(elapsed_time, log_main)
            self.shape_requests = []
            self._backup_dict = dict()
        else:
            raise TooManyDispersionTrialsError(f'Dispersion has been unsuccessful after {max_trials} trials.')

    def _log_dispersion_success(self, elapsed_time, log_main):
        """Log the success of the dispersion process."""
        success_msg = (f'{self.num_requested_shapes} shapes dispersed successfully '
                       f'in {elapsed_time:.2f} seconds.')
        self.dispersion_logger.debug(success_msg + '\n')
        if log_main:
            logger.info(success_msg)

    def _find_suitable_num_shapes(self, target_vf: float, vf_tolerance: float,
                                  min_num_shapes: int, max_num_shapes: int,
                                  solver_mode: str = 'BISECTION') -> int:
        """Find the suitable number of shapes for the *ShapeDispersionArray*'s requested shapes.

        This function first makes sure that all the shape requests are without *num_shapes*,
        then finds a suitable number of shapes that satisfies a *target_vf*.
        In the case of multiple requests, an equal number of shapes will be used for all of them.

        Read the documentation for the :meth:`_find_vf_for_shape_number` for more information.
        This is a private method used by the :meth:`disperse_shapes_vf` method
        and should not be directly called by the users.

        Args:
            target_vf: Target volume fraction.
            vf_tolerance: Maximum tolerable difference between the reached volume fraction and *target_vf*.
                          It should be a float greater than 1E-4.
            min_num_shapes: The initial value of *num_shapes* investigated by the function.
            max_num_shapes: The maximum value of *num_shapes* investigated by the function.
            solver_mode: The solver used for finding the suitable number of shapes.
                         If set to *'BRUTE FORCE'*, all values from *min_num_shapes* to *max_num_shapes*
                         are tested and if set to *'BISECTION'*, the bisection method
                         is used to find the suitable value.
                         Defaults to *'BISECTION'* which is faster and more reliable.

        Returns:
            A suitable value for *num_shapes*.

        Raises:
            SuitableNumShapesNotFoundError: If a suitable number of shapes is not found in the given range.
        """
        begin_log_msg = 'Trying to find a suitable number of shapes ' \
                        'for the given shape requests and target volume fraction.'
        self.dispersion_logger.debug(begin_log_msg + '\n')
        logger.info(begin_log_msg)

        # Validate input.
        if target_vf <= 0:
            raise ValueError('target_volume_fraction should be a positive number.')
        if (int(min_num_shapes) != min_num_shapes) or (min_num_shapes < 1):
            raise ValueError(f'min_num_shapes must be an integer greater than 1, but is {min_num_shapes}.')
        if (int(max_num_shapes) != max_num_shapes) or (max_num_shapes < 1):
            raise ValueError(f'max_num_shapes must be an integer greater than 2, but is {max_num_shapes}.')
        if max_num_shapes <= min_num_shapes:
            raise ValueError(f'max_num_shapes must be greater than min_num_shapes.')
        if vf_tolerance < 1E-4:
            raise ValueError('vf_tolerance should be a float greater than 1E-4.')
        if solver_mode.upper() not in ['BRUTE FORCE', 'BISECTION']:
            raise ValueError('Invalid value for solver_mode.')
        # Make sure all num_shapes in the shape requests are None.
        num_shapes_list = [sr[1] for sr in self.shape_requests]
        if not all(ns is None for ns in num_shapes_list):
            raise ValueError(f'Some of the num_shapes in shape_requests are *not* None.'
                             f'If you want to find suitable num_shapes, they should *all* start as None.'
                             f'The list is: {num_shapes_list}')

        with LogWithoutFormatContext(self.dispersion_logger):
            self.dispersion_logger.debug(f' Using the {solver_mode.title()} solver mode.\n')
        line_str = ('=' * 80) + '\n'
        if solver_mode.upper() == 'BRUTE FORCE':
            solver_function = integer_solver_brute_force
        elif solver_mode.upper() == 'BISECTION':
            solver_function = integer_solver_bisection
        else:
            raise RuntimeError('Invalid value for solver_mode. This should have been caught earlier.')

        num_shapes = solver_function(func=self._find_vf_for_shape_number,
                                     a=min_num_shapes, b=max_num_shapes, tolerance=vf_tolerance,
                                     target_vf=target_vf)
        self.dispersion_logger.debug(f'Suitable number of shapes was found '
                                     f'to be {num_shapes}x{len(self.shape_requests)}.\n' + line_str)
        return num_shapes

    def _find_vf_for_shape_number(self, num_shapes: int, target_vf: float) -> float:
        """Calculate the volume fraction (VF) for the *ShapeDispersionArray*'s requested shapes,
        given the input number of shapes and return the difference between the actual and target VF.

        This function checks if all the shape requests are without *num_shapes*,
        then creates the requested shapes in a list.
        In the case of multiple requests each one of them will be created in *equal* numbers.
        Afterwards, the analytical volume of each shape will be calculated
        using the shape's :attr:`~.BaseShape.analytical_volume` property
        and these volume will be added and converted to a total VF.
        Finally, the *target_vf* is subtracted from the actual VF and returned.

        Notes:
            - The shapes created by this function may not be necessarily
              dispersable in a ShapeDispersionArray. This function is only about the VFs.
            - The nature of shape generation is stochastic. This means that no two calls
              will have an equal output. This is generally OK as the results are expected
              within a tolerance.
            - The volume used is the analytical volume of each shape
              which is fast but inaccurate for larger voxel sizes.
            - The volume of each voxel is calculated individually.
              This means that overlapping shapes and contact with the boundary
              will cause different VFs. A different function needs to be used for those cases.

        Args:
            num_shapes: Number of shapes to be created and investigated by the function.
            target_vf: Target volume fraction.

        Returns:
            The difference between the actual VF and *target_vf*.
        """
        # This function is used many times and therefore logs nothing.
        # Validate input.
        if target_vf <= 0:
            raise ValueError('target_volume_fraction should be a positive number.')
        # Make sure all num_shapes in the shape requests are None.
        num_shapes_list = [sr[1] for sr in self.shape_requests]
        if not all(ns is None for ns in num_shapes_list):
            raise ValueError(f'Some of the num_shapes in shape_requests are *not* None.'
                             f'If you want to find suitable num_shapes, they should *all* start as None.'
                             f'The list is: {num_shapes_list}')

        shape_list = []  # Note that this is a simple list and not a ShapeArray instance.
        for sr in self.shape_requests:
            (cls, _, iterable_kwargs, scalar_kwargs) = sr
            # Regenerate the BaseNormalDistributionDispersion subclasses with num_shapes values.
            for k, v in iterable_kwargs.items():
                if isinstance(v, ManualListDispersion):
                    raise ValueError(f'{k} is a ManualListDispersion which is not allowed '
                                     'when finding num_shapes.')
                elif isinstance(v, BaseNormalDistributionDispersion):
                    v.generate_values(num_values=num_shapes)
                else:
                    raise RuntimeError(f'Unexpected type {type(v)} for {k}: {v}.')

            # Create the shapes as single independent instances in shape_list.
            for i in range(num_shapes):
                # Put the shape's kwargs into a single dict to be passed as scalars.
                shape_i_kwargs = dict()
                for k, v in iterable_kwargs.items():
                    shape_i_kwargs[k] = v[i]
                for k, v in scalar_kwargs.items():
                    if isinstance(v, RandomDispersion):
                        shape_i_kwargs[k] = v()
                    else:
                        shape_i_kwargs[k] = v

                # Create the shape.
                # Note that a shape instance is not placed anywhere and simply exists.
                # In fact, the coordinate variable should have a RandomDispersion instance
                # as the value which will raise an error if used.
                # But we only need to calculate the analytical volume so this is OK.
                shape_list.append(cls(id=i + 1, **shape_i_kwargs))

        # Calculate the sum of analytical volumes for the shapes in shape_list.
        total_analytical_volume = sum(i.analytical_volume for i in shape_list)
        # Calculate volume fraction and the difference between it and target_vf.
        current_vf = total_analytical_volume / self.part_volume
        vf_diff = current_vf - target_vf
        self.dispersion_logger.debug(
            f'Tried num_shapes={num_shapes:4d}x{len(self.shape_requests)}, '
            f'Total Vol={total_analytical_volume:10.6f}, '
            f'VF={current_vf:10.6f}, Target VF={target_vf:.6} -> '
            f'Delta VF={vf_diff:+11.6f}\n')
        return vf_diff

    def disperse_shapes_vf(self, target_vf: float, vf_tolerance: float,
                           min_num_shapes: int = 5, max_num_shapes: int = 5000,
                           solver_mode: str = 'BISECTION',
                           max_attempts: int = 5000, max_trials: int = 100,
                           max_generations: int = 100):
        """Disperse shapes in the *ShapeDispersionArray* instance to reach a target volume fraction (VF).
        All shape requests are processed and then :attr:`shape_requests` is emptied.

        Here, all shape requests should have a *num_shapes* value of *None*.
        The function uses the :meth:`_find_suitable_num_shapes` function to find
        a suitable number of shapes (for all requests).

        Then, for *max_generations* number of times, the following is performed:

          1. The random values in shape requests are regenerated.
          2. The :meth:`disperse_shapes` method is used to disperse the shapes.
             Make sure to read that function's documents for its intricacies.
             If the shapes cannot be dispersed, a new generation is started.
          3. If dispersion is successful, the final volume fraction (VF)
             is checked to make sure it is within tolerances.

        Args:
            target_vf: Target volume fraction.
            vf_tolerance: Maximum tolerable difference between the reached volume fraction and *target_vf*.
                          It should be a float greater than 1E-4.
            min_num_shapes: The initial value used to find a suitable *num_shapes*.
            max_num_shapes: The maximum value used for finding a suitable *num_shapes*.
            solver_mode: The solver used for finding the suitable number of shapes.
                         If set to *'BRUTE FORCE'*, all values from *min_num_shapes* to *max_num_shapes*
                         are tested and if set to *'BISECTION'*, the bisection method
                         is used to find the suitable value.
                         Defaults to *'BISECTION'* which is faster and more reliable.
            max_attempts: The maximum number of attempts for placement of a shape.
                          If exceeded, the trial ends and the process is restarted.
            max_trials:   The maximum number of trials for a shape generation attempt.
                          If exceeded, a new shape generation attempt will be made.
            max_generations: Maximum number of shape generation attempts.
                             If exceeded, :class:`TooManyDispersionGenerationsError` is raised.

        Raises:
            TooManyDispersionGenerationsError: Too many trials were done.
                                               Note that each generation entails many trials.
        """
        self.dispersion_logger.debug('Dispersing shapes in the ShapeDispersionArray.\n')
        logger.debug('Dispersing shapes in the ShapeDispersionArray.')
        # Try for max_generations times to find the suitable number of shapes.
        for i in range(max_generations):
            try:
                num_shapes = self._find_suitable_num_shapes(target_vf, vf_tolerance,
                                                            min_num_shapes, max_num_shapes, solver_mode)
                break
            except SuitableNumShapesNotFoundError:
                continue
        # Set num_shapes in all shape requests and regenerate subclasses of BaseNormalDistributionDispersion.
        for sr in self.shape_requests:
            # noinspection PyUnboundLocalVariable
            sr[1] = num_shapes

        self.dispersion_logger.debug(f"Trying to disperse shapes in part '{self.part_name}'.\n"
                                     f"{' ' * 11}Target volume fraction is {target_vf:.6} "
                                     f"and tolerance is {vf_tolerance}.\n"
                                     f"{' ' * 11}Shape requests will be regenerated a maximum of {10} times.\n"
                                     f"{' ' * 11}Each generation will be dispersed for {max_trials} tries,\n"
                                     f"{' ' * 11}in which {max_attempts} placement attempts "
                                     f"will be made for each shape.\n")
        logger.info(f"Trying to disperse shapes in part '{self.part_name}'. "
                    f"Target volume fraction is {target_vf:.6} and tolerance is {vf_tolerance}.")
        # Backup the shape array instance's state.
        # Note that the instance's _backup_dict will be overwritten
        # as part of the process, so we have to create a copy for this loop.
        self._backup_state()
        vf_dispersion_backup = deepcopy(self._backup_dict)
        for gen_number in range(1, max_generations + 1):
            # fix this. it doesn't loop properly.
            # Regenerate all shape requests (subclasses of BaseNormalDistributionDispersion).
            self.dispersion_logger.debug(f'** Generation {gen_number}: Regenerating shape requests ... ')
            for sr in self.shape_requests:
                iterable_kwargs = sr[2]
                for k, v in iterable_kwargs.items():
                    if isinstance(v, BaseNormalDistributionDispersion):
                        v.generate_values(num_values=sr[1])
            with LogWithoutFormatContext(self.dispersion_logger):
                self.dispersion_logger.debug('Done. **\n')

            # Try to disperse this generation of shapes in the structure.
            try:
                self.disperse_shapes(max_attempts=max_attempts, max_trials=max_trials, log_main=False)
            except TooManyDispersionTrialsError:
                self.dispersion_logger.debug(
                    f'Generation {gen_number} could not be dispersed '
                    f'after {max_trials} trials. Regenerating.\n')
                self._restore_state(backup_dict=vf_dispersion_backup)
                continue

            # Compare target VF with the instance's real voxel-based VF.
            vf_diff = abs(self.shape_array_volume_fraction - target_vf)
            if vf_diff <= vf_tolerance:
                vf_dispersion_status = True
                dispersion_status_str = '... Accepted!'
            else:
                vf_dispersion_status = False
                dispersion_status_str = '... Rejected!'

            self.dispersion_logger.debug(
                f"Generation {gen_number} dispersed. "
                f"VF={self.shape_array_volume_fraction:.6f}, "
                f"Target VF={target_vf:.6}, Delta VF={vf_diff:.6f} "
                f"{'<' if vf_dispersion_status else '>'} {vf_tolerance} "
                f"{dispersion_status_str}\n")

            if vf_dispersion_status:
                success_msg = (f'Dispersion based on volume fraction was successful '
                               f'within the given tolerance after {gen_number} generations.')
                self.dispersion_logger.debug(success_msg + '\n')
                logger.info(success_msg)
                return
            else:
                self._restore_state(backup_dict=vf_dispersion_backup)
                continue
        raise TooManyDispersionGenerationsError(f'Dispersion failed after {max_generations} generations.')


def integer_solver_validator(func: Callable, a: int, b: int, tolerance: float):
    """Validate inputs for the integer solver functions.

    Args:
        func: The continuous function :math:`F()` to be solved.
              This function only checks if it is callable.
        a: The start of the range where the search takes place. Should be an integer less than *b*.
        b: The end of the range where the search takes place. Should be an integer greater than *a*.
        tolerance: The tolerance for finding the root of the function. It should be greater than 1E-4.

    Raises:
        ValueError: If any of the input values are invalid.
    """
    if not callable(func):
        raise ValueError(f'func should be a function or a method but its type is {type(func)}')
    if (int(a) != a) or (a < 1):
        raise ValueError(f'a must be an integer greater than 1, but is {a}.')
    if (int(b) != b) or (b < 1):
        raise ValueError(f'b must be an integer greater than 1, but is {b}.')
    if b <= a:
        raise ValueError(f'b must be greater than a.')
    if tolerance < 1E-4:
        raise ValueError('tolerance should be a float greater than 1E-4.')


def integer_solver_brute_force(func: Callable, a: int, b: int, tolerance: float, **kwargs) -> int:
    """A solver for finding the root of the function :math:`F()` using a brute force approach.

    This solver checks all values between *a* and *b* for a root that satisfies the given *tolerance*.
    The main difference between this and a regular solver is that in here, everything is an integer.

    Note that since the function :math:`F()` is stochastic in nature, there will always be
    an error associated with the result. This is OK as the result is acceptable with a tolerance.

    Args:
        func: The continuous function :math:`F()` to be solved. This solver is meant for and tested with
              the :meth:`ShapeDispersionArray._find_vf_for_shape_number` function,
              but should work for other functions.
        a: The start of the range where the search takes place. Should be an integer less than *b*.
        b: The end of the range where the search takes place. Should be an integer greater than *a*.
        tolerance: The tolerance for finding the root of the function. It should be greater than 1E-4.
        kwargs: The keyword arguments to be passed to *func*.

    Returns:
        An integer root for the function :math:`F()`.

    Raises:
        SuitableNumShapesNotFoundError: If a suitable root is not found.
    """
    # Validate inputs.
    integer_solver_validator(func, a, b, tolerance)

    old_vf_diff = None
    for num_shapes in range(a, b + 1):
        vf_diff = func(num_shapes, **kwargs)
        # If within acceptable range, return here.
        if abs(vf_diff) <= tolerance:
            return num_shapes
        # If the sign of new and old vf_diff are different,
        # it means we are only getting further from a good solution
        # and the best two solutions on the boundary have already
        # been determined to be outside tolerance.
        # So we should break the loop and raise SuitableNumShapesNotFoundError.
        if np.sign(old_vf_diff) == np.sign(vf_diff):
            break
    # If we are here, all values have been checked or a solution cannot be found. Raise error.
    raise SuitableNumShapesNotFoundError(f'Suitable num_shapes not found in the range [{a},{b}].')


def integer_solver_bisection(func: Callable, a: int, b: int, tolerance: float, **kwargs) -> int:
    """A solver for finding the root of the function :math:`F()` using
    the `bisection method <https://en.wikipedia.org/wiki/Bisection_method>`_.

    The main difference between this and a regular solver is that in here, everything is an integer.
    This means that if :math:`a=b`, or if :math:`b-a<1` and :math:`F(a)` or :math:`F(b)`
    are within the tolerances, the algorithms stops.

    Note that since the function :math:`F()` is stochastic in nature, there will always be
    an error associated with the result. This is OK as the result is acceptable with a tolerance.

    Args:
        func: The continuous function :math:`F()` to be solved. This solver is meant for and tested with
              the :meth:`ShapeDispersionArray._find_vf_for_shape_number` function,
              but should work for other functions.
              Note that :math:`F(a)` and :math:`F(b)` should have opposing signs.
        a: The start of the range where the search takes place. Should be an integer less than *b*.
        b: The end of the range where the search takes place. Should be an integer greater than *a*.
        tolerance: The tolerance for finding the root of the function. It should be greater than 1E-4.
        kwargs: The keyword arguments to be passed to *func*.

    Returns:
        An integer root for the function :math:`F()`.

    Raises:
        SuitableNumShapesNotFoundError: If a suitable root is not found.
    """
    # Validate inputs.
    integer_solver_validator(func, a, b, tolerance)

    f_a = func(a, **kwargs)
    f_b = func(b, **kwargs)
    if (f_a * f_b) > 0:
        raise ValueError('func(a) and func(b) must have opposite values. Try changing a and b.')

    while True:
        if abs(b - a) <= 1:
            if a == b:
                return a
            elif abs(f_a) < tolerance:
                return a
            elif abs(f_b) < tolerance:
                return b
            else:
                raise SuitableNumShapesNotFoundError(
                    f'The solver reached F(a={a})={f_a:.6} and F(b={b})={f_b:.6}, '
                    f'but could not find a root with the given tolerance ({tolerance:.6}).')
        # Calculate the *integer* midpoint c. Note that it will not be equal to a or b because of the above check.
        c = int((a + b) / 2)
        f_c = func(c, **kwargs)
        if abs(f_c) < tolerance:  # F(c) is OK.
            return c
        elif f_a * f_c < 0:  # The root is between a and c.
            b = c
            f_b = f_c
        else:  # The root is between c and b.
            a = c
            f_a = f_c



"""Functions used for creating boundary conditions.
The BC and it's information is stored in
the :class:`~vcams.voxelpart.VoxelPart` object and written to the output with it.

See the :ref:`boundary-conditions` section for a complete explanation
of the basic concepts.
"""

import logging
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Union, Iterable

from numpy import ravel_multi_index, array, arange, meshgrid, concatenate, isin, count_nonzero
from numpy.typing import NDArray

from vcams.helper import validate_materials_to_be_output

logger = logging.getLogger(__name__)


class BcNotApplicableError(Exception):
    pass


@dataclass
class TieConstraint:
    """Class for tying a master set (single node) and a slave set (multiple nodes).
    The tie is defined using the ``*EQUATION`` keyword with the coefficients set to +1 and -1."""
    dof: int
    """Degree of freedom used in the constraint."""
    rp_set_name: str
    """Name of the set containing only a single master node (AKA dummy node/reference point)."""
    slave_set_name: str
    """Name of the set containing the slave nodes."""

    def __repr__(self):
        return (f'*Equation\n2\n'
                f'"{self.slave_set_name}", {self.dof}, -1.\n'
                f'"{self.rp_set_name}", {self.dof}, +1.\n')


@dataclass
class BasePbcConstraint(ABC):
    """Abstract base class for defining a constraint for a Periodic Boundary Condition (PBC).

    Subclasses are created for 2D and 3D spaces and for faces, edges, and vertices.
    This dataclass simply defines the required parameters and establishes a common interface.
    tying a master set (single node) and a slave set (multiple nodes).
    The tie is defined using the ``*EQUATION`` keyword with the coefficients set to +1 and -1.

    Subclasses must redefine the ``__repr__()`` method to output
    a complete statement based on the ``*EQUATION`` keyword
    of *Abaqus Keywords Reference Guide*.
    """
    part_instance_name: str
    """Name of the part instance that the nodes belong to."""
    dummy_names: tuple[str]
    """Name of the set containing the dummy node, or a tuple of the names."""
    dummy_coeffs: tuple[float]
    """Value of the coefficient for the dummy node, or a tuple of the coefficients."""
    node1_id: int
    """ID of the first node used for the equation."""
    node2_id: int
    """ID of the second node used for the equation."""

    @abstractmethod
    def __repr__(self):
        pass


@dataclass
class Pbc3DFaceConstraint(BasePbcConstraint):
    """Class for defining the following 3D PBC constraint
    between a dummy nodes and two corresponding nodes on the faces of a cuboid:

    .. math::
       u_i^{F_{j2}} - u_i^{F_{j1}} + C_j u_i^{D_j} = 0

    where :math:`F_{j1}` and :math:`F_{j2}` are the two faces in the :math:`j` direction,
    :math:`D_j` and :math:`C_j` are the dummy node and the coefficient used for that direction,
    and the equation is written for all DOFs, i.e. :math:`i=1,2,3`
    Note that :math:`C_j` is typically supplied as a negative value.

    The parameters for creating an object are similar to :class:`BasePbcConstraint`,
    except for the meaning of *dummy_names* and *dummy_coeffs* which are:
    """
    dummy_names: str
    """Name of the set containing the dummy node :math:`D_j`."""
    dummy_coeffs: float
    """Value of the coefficient :math:`C_j`."""

    def __repr__(self):
        return ''.join((f'*Equation\n3\n'
                        f'"{self.part_instance_name}".{self.node2_id + 1}, {dof}, 1.\n'
                        f'"{self.part_instance_name}".{self.node1_id + 1}, {dof}, -1.\n'
                        f'"{self.dummy_names}", {dof}, {self.dummy_coeffs}\n')
                       for dof in (1, 2, 3))


@dataclass
class Pbc3DEdgeConstraint(BasePbcConstraint):
    """Class for defining the following 3D PBC constraint
    between a dummy nodes and two corresponding nodes on the edges of a cuboid:

    .. math::
       u_i^{E2} - u_i^{E1} + C_1 u_i^{D_1} + C_2 u_i^{D_2} = 0

    where :math:`E1` and :math:`E2` are two compatible edges,
    :math:`D_1` and :math:`D_2` are the dummy nodes used for the edges,
    :math:`C_1` and :math:`C_2` are the coefficients used those dummy nodes,
    and the equation is written for all DOFs, i.e. :math:`i=1,2,3`.

    The above equation is a generalization of the edge equations
    in Eq. :eq:`bc-eq-pbc`. Compatible edges, dummy nodes,
    and the value of coefficients must be taken from Eq. :eq:`bc-eq-pbc`.

    The parameters for creating an object are similar to :class:`BasePbcConstraint`.
    """

    def __repr__(self):
        return ''.join((f'*Equation\n4\n'
                        f'"{self.part_instance_name}".{self.node2_id + 1}, {dof}, 1.\n'
                        f'"{self.part_instance_name}".{self.node1_id + 1}, {dof}, -1.\n'
                        f'"{self.dummy_names[0]}", {dof}, {self.dummy_coeffs[0]}\n'
                        f'"{self.dummy_names[1]}", {dof}, {self.dummy_coeffs[1]}\n')
                       for dof in (1, 2, 3))


class Pbc3DVertexConstraint(BasePbcConstraint):
    """Class for defining the following 3D PBC constraint
    between a dummy nodes and two nodes on vertices of a cuboid:

    .. math::
       u_i^{V2} - u_i^{V1} + C_1 u_i^{D_1} + C_2 u_i^{D_2} + C_3 u_i^{D_3} = 0

    where :math:`V1` and :math:`V2` are two compatible vertices,
    :math:`D_j` and :math:`C_j` are the dummy nodes and coefficients
    (all are used), and the equation is written for all DOFs, i.e. :math:`i=1,2,3`.

    The above equation is a generalization of the vertex equations
    in Eq. :eq:`bc-eq-pbc`. Compatible vertices and the value of coefficients
    must be taken from Eq. :eq:`bc-eq-pbc`.

    The parameters for creating an object are similar to :class:`BasePbcConstraint`.
    """

    def __repr__(self):
        return ''.join((f'*Equation\n5\n'
                        f'"{self.part_instance_name}".{self.node2_id + 1}, {dof}, 1.\n'
                        f'"{self.part_instance_name}".{self.node1_id + 1}, {dof}, -1.\n'
                        f'"{self.dummy_names[0]}", {dof}, {self.dummy_coeffs[0]}\n'
                        f'"{self.dummy_names[1]}", {dof}, {self.dummy_coeffs[1]}\n'
                        f'"{self.dummy_names[2]}", {dof}, {self.dummy_coeffs[2]}\n')
                       for dof in (1, 2, 3))


@dataclass
class Pbc2DEdgeConstraint:
    """Class for defining the following *two* 2D PBC constraints
    between a dummy nodes and two nodes on vertices of a square:

    .. math::
       u_{dof}^{V2} - u_{dof}^{V1} + C u_{dof}^{D} = 0 \\\\
       u_{aux\\_dof}^{V2} - u_{aux\\_dof}^{V1} = 0

    where :math:`V1` and :math:`V2` are two compatible vertices,
    :math:`D` and :math:`C` are the dummy node and coefficient
    used for the equation.

    Note that this class behaves differently from the children of :class:`BasePbcConstraint`.
    While they define equations for all DOFs, this class receives a *dof* which can
    either be 1 or 2, and sets *aux_dof* to be the other one.

    The above equation is a generalization of the vertex equations
    in Eq. :eq:`bc-eq-pbc2d`. Compatible vertices and the value of coefficients
    must be taken from Eq. :eq:`bc-eq-pbc2d`.

    The parameters for creating an object are:
    """
    part_instance_name: str
    """Name of the part instance that the nodes belong to."""
    dof: int
    """The DOF for which the first equation is defined.
    It must be 1 or 2 and the other value is assigned to *aux_dof*
    for which the second equation is defined."""
    dummy_names: str
    """Name of the set containing the dummy node."""
    dummy_coeffs: float
    """Value of the coefficient for the dummy node."""
    node1_id: int
    """ID of the first node used for the equation."""
    node2_id: int
    """ID of the second node used for the equation."""

    def __post_init__(self):
        if self.dof == 1:
            self.aux_dof: int = 2
        elif self.dof == 2:
            self.aux_dof: int = 1
        else:
            raise ValueError('Invalid value for dof. Only 1 and 2 are allowed.')

    def __repr__(self):
        return (f'*Equation\n3\n'
                f'"{self.part_instance_name}".{self.node2_id + 1}, {self.dof}, 1.\n'
                f'"{self.part_instance_name}".{self.node1_id + 1}, {self.dof}, -1.\n'
                f'"{self.dummy_names}", {self.dof}, {self.dummy_coeffs}\n'
                f'*Equation\n2\n'
                f'"{self.part_instance_name}".{self.node2_id + 1}, {self.aux_dof}, 1.\n'
                f'"{self.part_instance_name}".{self.node1_id + 1}, {self.aux_dof}, -1.\n')


class Pbc2DVertexConstraint(BasePbcConstraint):
    """Class for defining the following 2D PBC constraint
    between a dummy nodes and two nodes on vertices of a square:

    .. math::
       u_i^{V2} - u_i^{V1} + C_1 u_i^{D_1} + C_2 u_i^{D_2} = 0

    where :math:`V1` and :math:`V2` are two compatible vertices,
    :math:`D_j` and :math:`C_j` are the dummy nodes and coefficients
    (all are used), and the equation is written for all DOFs, i.e. :math:`i=1,2`.

    The above equation is a generalization of the vertex equations
    in Eq. :eq:`bc-eq-pbc2d`. Compatible vertices and the value of coefficients
    must be taken from Eq. :eq:`bc-eq-pbc2d`.

    The parameters for creating an object are similar to :class:`BasePbcConstraint`.
    """

    def __repr__(self):
        return ''.join((f'*Equation\n4\n'
                        f'"{self.part_instance_name}".{self.node2_id + 1}, {dof}, 1.\n'
                        f'"{self.part_instance_name}".{self.node1_id + 1}, {dof}, -1.\n'
                        f'"{self.dummy_names[0]}", {dof}, {self.dummy_coeffs[0]}\n'
                        f'"{self.dummy_names[1]}", {dof}, {self.dummy_coeffs[1]}\n')
                       for dof in (1, 2))


# noinspection PyProtectedMember
def create_bc(part, dim: str, mat_codes_to_accept: str | int | Iterable | NDArray,
              max_empty_border_elems: float = 0.5) -> list:
    """Create a boundary condition (BC) for a VoxelPart object.

    This function uses the VoxelPart object's *_bc_type* property and
    (1) creates the relevant node sets using :func:`create_node_sets`
    and, (2) returns a tuple of constraint objects to be written to output.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` object on which the operation is performed.
        dim: Dimensionality of the intended output. Valid values are '2D' and '3D'.
        max_empty_border_elems: Maximum ratio of empty elements in each border or edge.
                                Defaults to 0.5 which means 50%.
        mat_codes_to_accept: See the *material_list* parameter of
                             :func:`helper.validate_materials_to_be_output`.

    Returns:
        A list of constraint objects. The number and class of list contents
        depends on the VoxelPart object's *_bc_type* property.
    """
    if dim.upper() not in ['2D', '3D']:
        raise ValueError("dim can only be one of '2D' or '3D'.")

    bc_type = part._bc_type
    if bc_type is None:
        part._bc_add_dummy_nodes = False
        logger.info('No BCs have been created.')
        return []

    bc_applicability_status = check_border_elements(part, dim,
                                                    mat_codes_to_accept, max_empty_border_elems)
    if bc_applicability_status == 'OK':
        pass
    elif bc_applicability_status == 'WINDOW':
        raise BcNotApplicableError("The part's boundaries contain some empty elements which invalidate BCs. "
                                   "This can be mitigated using The Window Method,"
                                   "which has not been implemented yet.")
    elif bc_applicability_status == 'UNABLE':
        raise BcNotApplicableError("The part's boundaries contain too many empty elements which invalidate BCs.")

    if bc_type.upper() == 'NODESET ONLY':
        create_node_sets(part, dim, explicit_sets=part._bc_nodeset_explicit,
                         simple_sets=part._bc_nodeset_simple)
        logger.info('No BCs have been created but node sets were added to the part.')
        return []

    elif bc_type.upper() == 'LINEAR DISPLACEMENT':
        part._add_dummy_nodes(fixed=True, single_node=True)
        create_node_sets(part, dim, explicit_sets=part._bc_nodeset_explicit, simple_sets=True)
        if dim.upper() == '2D':
            constraint_list = [TieConstraint(dof=1, rp_set_name='RP0-NodeSet', slave_set_name='Simple-Edge11-NodeSet'),
                               TieConstraint(dof=2, rp_set_name='RP0-NodeSet', slave_set_name='Simple-Edge21-NodeSet'),
                               TieConstraint(dof=1, rp_set_name='RP1-NodeSet', slave_set_name='Simple-Edge12-NodeSet'),
                               TieConstraint(dof=2, rp_set_name='RP1-NodeSet', slave_set_name='Simple-Edge22-NodeSet')]
        else:  # dim.upper() == '3D'
            constraint_list = [TieConstraint(dof=1, rp_set_name='RP0-NodeSet', slave_set_name='Simple-Face11-NodeSet'),
                               TieConstraint(dof=2, rp_set_name='RP0-NodeSet', slave_set_name='Simple-Face21-NodeSet'),
                               TieConstraint(dof=3, rp_set_name='RP0-NodeSet', slave_set_name='Simple-Face31-NodeSet'),
                               TieConstraint(dof=1, rp_set_name='RP1-NodeSet', slave_set_name='Simple-Face12-NodeSet'),
                               TieConstraint(dof=2, rp_set_name='RP1-NodeSet', slave_set_name='Simple-Face22-NodeSet'),
                               TieConstraint(dof=3, rp_set_name='RP1-NodeSet', slave_set_name='Simple-Face32-NodeSet')]

        logger.info('Linear Displacement BCs have were created.')
        return constraint_list

    elif bc_type.upper() == 'PERIODIC':
        part._add_dummy_nodes(fixed=False, three_nodes=True)
        create_node_sets(part, dim, explicit_sets=True, simple_sets=part._bc_nodeset_simple)
        constraint_list = []
        pl = part.real_size  # Length of the part.

        if dim.upper() == '2D':
            # Add constraints for the edges.
            constraint_list += add_2d_pbc_constraints(part, 'edge', dof=2,
                                                      dummy_names='RP2-NodeSet', dummy_coeffs=-pl[1],
                                                      set_names=('Edge12-NodeSet', 'Edge34-NodeSet'))
            constraint_list += add_2d_pbc_constraints(part, 'edge', dof=1,
                                                      dummy_names='RP1-NodeSet', dummy_coeffs=-pl[0],
                                                      set_names=('Edge14-NodeSet', 'Edge23-NodeSet'))
            # Add constraints for the vertices.
            constraint_list += add_2d_pbc_constraints(part, 'vertex', dof=None,
                                                      dummy_names=('RP1-NodeSet', 'RP2-NodeSet'),
                                                      dummy_coeffs=(-pl[0], -pl[1]),
                                                      set_names=('Vertex1-NodeSet', 'Vertex3-NodeSet'))
            constraint_list += add_2d_pbc_constraints(part, 'vertex', dof=None,
                                                      dummy_names=('RP1-NodeSet', 'RP2-NodeSet'),
                                                      dummy_coeffs=(+pl[0], -pl[1]),
                                                      set_names=('Vertex2-NodeSet', 'Vertex4-NodeSet'))
        else:  # dim.upper() == '3D'
            # Add constraints for the faces.
            constraint_list += add_3d_pbc_constraints(part, 'face', 'RP1-NodeSet', -pl[0],
                                                      ('Face11-NodeSet', 'Face12-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'face', 'RP2-NodeSet', -pl[1],
                                                      ('Face21-NodeSet', 'Face22-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'face', 'RP3-NodeSet', -pl[2],
                                                      ('Face31-NodeSet', 'Face32-NodeSet'))
            # Add constraints for the edges.
            constraint_list += add_3d_pbc_constraints(part, 'edge', ('RP2-NodeSet', 'RP3-NodeSet'), (-pl[1], -pl[2]),
                                                      ('Edge12-NodeSet', 'Edge78-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'edge', ('RP2-NodeSet', 'RP3-NodeSet'), (+pl[1], -pl[2]),
                                                      ('Edge34-NodeSet', 'Edge56-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'edge', ('RP1-NodeSet', 'RP2-NodeSet'), (-pl[0], -pl[1]),
                                                      ('Edge15-NodeSet', 'Edge37-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'edge', ('RP1-NodeSet', 'RP2-NodeSet'), (-pl[0], +pl[1]),
                                                      ('Edge48-NodeSet', 'Edge26-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'edge', ('RP1-NodeSet', 'RP3-NodeSet'), (-pl[0], -pl[2]),
                                                      ('Edge14-NodeSet', 'Edge67-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'edge', ('RP1-NodeSet', 'RP3-NodeSet'), (-pl[0], +pl[2]),
                                                      ('Edge58-NodeSet', 'Edge23-NodeSet'))
            # Add constraints for the vertices.
            dummy_names = ('RP1-NodeSet', 'RP2-NodeSet', 'RP3-NodeSet')
            constraint_list += add_3d_pbc_constraints(part, 'vertex', dummy_names, (-pl[0], -pl[1], -pl[2]),
                                                      ('Vertex1-NodeSet', 'Vertex7-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'vertex', dummy_names, (+pl[0], -pl[1], -pl[2]),
                                                      ('Vertex2-NodeSet', 'Vertex8-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'vertex', dummy_names, (+pl[0], +pl[1], -pl[2]),
                                                      ('Vertex3-NodeSet', 'Vertex5-NodeSet'))
            constraint_list += add_3d_pbc_constraints(part, 'vertex', dummy_names, (-pl[0], +pl[1], -pl[2]),
                                                      ('Vertex4-NodeSet', 'Vertex6-NodeSet'))
        logger.info('Periodic BCs have were created.')
        return constraint_list

    else:
        raise ValueError('Invalid value for bc_type.')


def check_border_elements(part, dim: str,
                          mat_codes_to_accept: str | int | Iterable | NDArray,
                          max_empty_border_elems: float = 0.5) -> str:
    """Check the border elements for empty space and return the status.

    Args:
        part (VoxelPart): The *VoxelPart* instance where operations take place.
        dim: Dimensionality of the intended output. Valid values are '2D' and '3D'.
        max_empty_border_elems: Maximum ratio of empty elements in each border or edge.
                                Defaults to 0.5 which means 50%.
        mat_codes_to_accept: See the *material_list* parameter of
                             :func:`helper.validate_materials_to_be_output`.

    Returns:
        One of the following strings

        - *'OK'* if all border elements are non-empty and the BC can be applied as is.
        - *'WINDOW'* if there are empty elements in some border areas,
          but the ratio of empty to non-empty is acceptable for each border area.
          This means that BCs can be applied using the Window Method.
        - *'FAIL'* if the part has too many non-empty elements in the borders and
          BCs cannot be applied.

    """
    # Validate parameters.
    mat_codes_to_accept = validate_materials_to_be_output(part, mat_codes_to_accept)
    if dim.upper() not in ['2D', '3D']:
        raise ValueError("dim can only be one of '2D' or '3D'.")
    if max_empty_border_elems <= 0.05:
        raise ValueError('max_empty_border_elems should be greater or equal to 0.05.')

    arr = part.data
    # Create a list of arrays for each border area.
    # Note that there is an overlap between border areas which is OK.
    if dim.upper() == '2D':
        border_arrays_list = (arr[0, :], arr[-1, :],
                              arr[:, 0], arr[:, -1])
    elif dim.upper() == '3D':
        border_arrays_list = (arr[0, :, :], arr[-1, :, :],
                              arr[:, 0, :], arr[:, -1, :],
                              arr[:, :, 0], arr[:, :, -1])
    else:
        raise RuntimeError('Invalid Value for dim. This should have been caught earlier.')

    # Count the number of rejected elements in each border array.
    num_rejected_elems_list = []
    num_border_elems_list = []
    for bal in border_arrays_list:
        num_rejected_elems_list.append(
            count_nonzero(
                isin(element=bal, test_elements=mat_codes_to_accept, invert=True, kind='table')))
        num_border_elems_list.append(bal.size)

    # Determine status based on the number of rejected elements.
    # Start with OK.
    status = 'OK'
    # If there are any rejected border elements the window method must be used.
    if any(array(num_rejected_elems_list) != 0):
        status = 'WINDOW'
    # If there are too many rejected border elements, even the window method won't work.
    if any((array(num_rejected_elems_list) / array(num_border_elems_list)) > max_empty_border_elems):
        status = 'UNABLE'

    return status


def add_2d_pbc_constraints(part, typ: str, dof: int | None,
                           dummy_names: Union[str, tuple],
                           dummy_coeffs: Union[float, tuple],
                           set_names: tuple) -> list:
    """Add an equation for a 2D PBC.
    With the correct arguments, all the equations
    in Eq. :eq:`bc-eq-pbc2d` can be implemented using this function.

    Args:
        part (VoxelPart): The *VoxelPart* object on which the operation is performed.
        typ: Type of the node sets to be constrained. Valid values are 'edge' and 'vertex'.
        dof: If *typ=edge*, must be either be 1 or 2, otherwise is not used. Defaults to *None*.
        dummy_names: If *typ=edge*, a tuple containing the names of the
                     two sets containing the dummy nodes for the equation.
                     If *typ=vertex*, a single string denoting the name of the vertex set.
                     Otherwise, an error is raised.
        dummy_coeffs: If *typ=edge*, a single float or
                      if *typ=vertex*, a tuple containing the coefficients for the dummy nodes.
                      Otherwise, an error is raised.
        set_names: Tuple of the names of the sets of the edges or vertices to be constrained.
                   The sets must contain the same number of nodes,
                   and they must be in the same order.

    Returns:
        A list of constraint objects for the given parameters.
    """
    set1_ids = part.node_sets[set_names[0]]
    set2_ids = part.node_sets[set_names[1]]
    instance_name = part.instance_name
    if len(set1_ids) != len(set2_ids):
        raise ValueError("Sets '%s' and '%s' are of different lengths."
                         "This should have been caught before calling this function.")
    if typ.lower() == 'edge':
        if dof not in (1, 2):
            raise ValueError(f"For type='edge', dof must be 1 or 2 but is {dof}.")
        return [Pbc2DEdgeConstraint(instance_name, dof, dummy_names, dummy_coeffs, set1_ids[i], set2_ids[i])
                for i in range(len(set1_ids))]
    elif typ.lower() == 'vertex':
        return [Pbc2DVertexConstraint(instance_name, dummy_names, dummy_coeffs, set1_ids[i], set2_ids[i])
                for i in range(len(set1_ids))]
    else:
        raise ValueError("typ must be either 'edge' or 'vertex'.")


def add_3d_pbc_constraints(part, typ: str,
                           dummy_names: Union[str, tuple],
                           dummy_coeffs: Union[float, tuple],
                           set_names: tuple) -> list:
    """Add an equation for a 3D PBC.
    With the correct arguments, all the equations
    in Eq. :eq:`bc-eq-pbc` can be implemented using this function.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` object on which the operation is performed.
        typ: Type of the node sets to be constrained. Valid values are 'face', 'edge', and 'vertex'.
        dummy_names: Tuple of the names of the sets containing the dummy nodes for the equation
                     or in the case of ``typ==vertex``, a single string.
        dummy_coeffs: Tuple of the values of the coefficients for the dummy nodes.
                      or in the case of ``typ==vertex``, a single float.
        set_names: Tuple of the names of the sets of the edges or vertices to be constrained.
                   The sets must contain the same number of nodes,
                   and they must be in the same order.

    Returns:
        A list of constraint objects for the given parameters.
    """
    set1_ids = part.node_sets[set_names[0]]
    set2_ids = part.node_sets[set_names[1]]
    instance_name = part.instance_name
    if len(set1_ids) != len(set2_ids):
        raise ValueError("Sets '%s' and '%s' are of different lengths."
                         "This should have been caught before calling this function.")
    if typ.lower() == 'face':
        constraint_class = Pbc3DFaceConstraint
    elif typ.lower() == 'edge':
        constraint_class = Pbc3DEdgeConstraint
    elif typ.lower() == 'vertex':
        constraint_class = Pbc3DVertexConstraint
    else:
        raise ValueError("typ must be one of 'face', 'edge', or 'vertex'.")
    return [constraint_class(instance_name, dummy_names, dummy_coeffs, set1_ids[i], set2_ids[i])
            for i in range(len(set1_ids))]


def create_node_sets(part, dim: str, explicit_sets: bool = False, simple_sets: bool = True):
    """Define node sets in a VoxelPart. They are created according to :numref:`bc-nodesets`:

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` object on which the operation is performed.
        dim: Dimensionality of the output part. Valid values are '2D' and '3D'.
        explicit_sets: If True, explicit node sets are created for vertices, edges, and faces.
                       as described in the section titled :ref:`boundary-conditions-pbc`.
                       Defaults to *False*.
        simple_sets: If True, simplified node sets are created
                     which are the full faces for the 3D models or the edges for the 2D models,
                     as described in the section titled :ref:`boundary-conditions-lin-disp`.
                     Defaults to *True*.
    """
    if dim.upper() not in ['2D', '3D']:
        raise ValueError("dim can only be one of '2D' or '3D'.")

    if not any([explicit_sets, simple_sets]):
        raise ValueError('At least one of explicit_sets and simple_sets must be set to True.')

    # Define a ravel function based on numpy.ravel_multi_index.
    def custom_ravel(inds):
        return ravel_multi_index(multi_index=inds, dims=node_array_shape,
                                 mode='raise', order='C').flatten()

    # Define a numpy.array containing a single number 0.
    zro = array([0])
    # Define elem_array_shape.
    elem_array_shape = part.size
    # In each direction, node array is larger by one.
    node_array_shape = tuple(i + 1 for i in elem_array_shape)
    if dim.upper() == '2D':
        max_x, max_y = elem_array_shape
    else:
        max_x, max_y, max_z = elem_array_shape
        inds_dir3 = arange(1, max_z)
    # Find indices of nodes in all directions. Indices for vertices are not included.
    inds_dir1 = arange(1, max_x)
    inds_dir2 = arange(1, max_y)

    # Define dictionaries for the sets.
    vertex_dict = dict()
    edge_dict = dict()
    face_dict = dict()
    simple_sets_dict = dict()

    # Define vertices.
    # Find IDs for vertices.
    if dim.upper() == '2D':
        vertex_coords = (array([0, max_x, max_x, 0]),
                         array([0, 0, max_y, max_y]))
    else:
        vertex_coords = (array([0, max_x, max_x, 0, 0, max_x, max_x, 0]),
                         array([0, 0, max_y, max_y, 0, 0, max_y, max_y]),
                         array([0, 0, 0, 0, max_z, max_z, max_z, max_z]))  # noqa F823
    # Find IDs and define node sets for the vertices.
    vertex_ids = custom_ravel(vertex_coords)
    for i in range(len(vertex_ids)):
        vertex_dict['Vertex%i-NodeSet' % (i + 1)] = vertex_ids[i]

    # Define edges.
    # Find IDs for edges. edge_ij refers to the edge formed by vertices i and j.
    if dim.upper() == '2D':
        edge_dict['Edge12-NodeSet'] = custom_ravel((inds_dir1, zro))
        edge_dict['Edge23-NodeSet'] = custom_ravel((max_x, inds_dir2))
        edge_dict['Edge34-NodeSet'] = custom_ravel((inds_dir1, max_y))
        edge_dict['Edge14-NodeSet'] = custom_ravel((zro, inds_dir2))
    else:
        edge_dict['Edge12-NodeSet'] = custom_ravel((inds_dir1, zro, zro))
        edge_dict['Edge23-NodeSet'] = custom_ravel((max_x, inds_dir2, zro))
        edge_dict['Edge34-NodeSet'] = custom_ravel((inds_dir1, max_y, zro))
        edge_dict['Edge14-NodeSet'] = custom_ravel((zro, inds_dir2, zro))
        edge_dict['Edge56-NodeSet'] = custom_ravel((inds_dir1, zro, max_z))
        edge_dict['Edge67-NodeSet'] = custom_ravel((max_x, inds_dir2, max_z))
        edge_dict['Edge78-NodeSet'] = custom_ravel((inds_dir1, max_y, max_z))
        edge_dict['Edge58-NodeSet'] = custom_ravel((zro, inds_dir2, max_z))
        edge_dict['Edge15-NodeSet'] = custom_ravel((zro, zro, inds_dir3))  # noqa F823
        edge_dict['Edge26-NodeSet'] = custom_ravel((max_x, zro, inds_dir3))
        edge_dict['Edge37-NodeSet'] = custom_ravel((max_x, max_y, inds_dir3))
        edge_dict['Edge48-NodeSet'] = custom_ravel((zro, max_y, inds_dir3))

    if dim.upper() == '3D':
        # Define faces.
        # Find IDs for faces. For face_ij, i refers the direction of normal vector
        # and j refers to whether this is the first or second edge in the direction.
        face_dict['Face11-NodeSet'] = custom_ravel(
            meshgrid(zro, inds_dir2, inds_dir3, sparse=True))
        face_dict['Face12-NodeSet'] = custom_ravel(
            meshgrid(max_x, inds_dir2, inds_dir3, sparse=True))
        face_dict['Face21-NodeSet'] = custom_ravel(
            meshgrid(inds_dir1, zro, inds_dir3, sparse=True))
        face_dict['Face22-NodeSet'] = custom_ravel(
            meshgrid(inds_dir1, max_y, inds_dir3, sparse=True))
        face_dict['Face31-NodeSet'] = custom_ravel(
            meshgrid(inds_dir1, inds_dir2, zro, sparse=True))
        face_dict['Face32-NodeSet'] = custom_ravel(
            meshgrid(inds_dir1, inds_dir2, max_z, sparse=True))

    if simple_sets:
        if dim == '2D':
            simple_sets_dict['Simple-Edge11-NodeSet'] = concatenate((edge_dict['Edge14-NodeSet'],
                                                                     array((vertex_dict['Vertex1-NodeSet'],
                                                                            vertex_dict['Vertex4-NodeSet']))),
                                                                    axis=0)
            simple_sets_dict['Simple-Edge12-NodeSet'] = concatenate((edge_dict['Edge23-NodeSet'],
                                                                     array((vertex_dict['Vertex2-NodeSet'],
                                                                            vertex_dict['Vertex3-NodeSet']))),
                                                                    axis=0)
            simple_sets_dict['Simple-Edge21-NodeSet'] = concatenate((edge_dict['Edge12-NodeSet'],
                                                                     array((vertex_dict['Vertex1-NodeSet'],
                                                                            vertex_dict['Vertex2-NodeSet']))),
                                                                    axis=0)
            simple_sets_dict['Simple-Edge22-NodeSet'] = concatenate((edge_dict['Edge34-NodeSet'],
                                                                     array((vertex_dict['Vertex3-NodeSet'],
                                                                            vertex_dict['Vertex4-NodeSet']))),
                                                                    axis=0)
        else:  # dim == '3D'
            simple_sets_dict['Simple-Face11-NodeSet'] = concatenate((face_dict['Face11-NodeSet'],
                                                                     edge_dict['Edge14-NodeSet'],
                                                                     edge_dict['Edge48-NodeSet'],
                                                                     edge_dict['Edge58-NodeSet'],
                                                                     edge_dict['Edge15-NodeSet'],
                                                                     array((vertex_dict['Vertex1-NodeSet'],
                                                                            vertex_dict['Vertex4-NodeSet'],
                                                                            vertex_dict['Vertex8-NodeSet'],
                                                                            vertex_dict['Vertex5-NodeSet']))),
                                                                    axis=None)
            simple_sets_dict['Simple-Face12-NodeSet'] = concatenate((face_dict['Face12-NodeSet'],
                                                                     edge_dict['Edge23-NodeSet'],
                                                                     edge_dict['Edge37-NodeSet'],
                                                                     edge_dict['Edge67-NodeSet'],
                                                                     edge_dict['Edge26-NodeSet'],
                                                                     array((vertex_dict['Vertex2-NodeSet'],
                                                                            vertex_dict['Vertex3-NodeSet'],
                                                                            vertex_dict['Vertex7-NodeSet'],
                                                                            vertex_dict['Vertex6-NodeSet']))),
                                                                    axis=None)
            simple_sets_dict['Simple-Face21-NodeSet'] = concatenate((face_dict['Face21-NodeSet'],
                                                                     edge_dict['Edge12-NodeSet'],
                                                                     edge_dict['Edge26-NodeSet'],
                                                                     edge_dict['Edge56-NodeSet'],
                                                                     edge_dict['Edge15-NodeSet'],
                                                                     array((vertex_dict['Vertex1-NodeSet'],
                                                                            vertex_dict['Vertex2-NodeSet'],
                                                                            vertex_dict['Vertex6-NodeSet'],
                                                                            vertex_dict['Vertex5-NodeSet']))),
                                                                    axis=None)
            simple_sets_dict['Simple-Face22-NodeSet'] = concatenate((face_dict['Face22-NodeSet'],
                                                                     edge_dict['Edge34-NodeSet'],
                                                                     edge_dict['Edge37-NodeSet'],
                                                                     edge_dict['Edge78-NodeSet'],
                                                                     edge_dict['Edge48-NodeSet'],
                                                                     array((vertex_dict['Vertex4-NodeSet'],
                                                                            vertex_dict['Vertex3-NodeSet'],
                                                                            vertex_dict['Vertex7-NodeSet'],
                                                                            vertex_dict['Vertex8-NodeSet']))),
                                                                    axis=None)
            simple_sets_dict['Simple-Face31-NodeSet'] = concatenate((face_dict['Face31-NodeSet'],
                                                                     edge_dict['Edge12-NodeSet'],
                                                                     edge_dict['Edge23-NodeSet'],
                                                                     edge_dict['Edge34-NodeSet'],
                                                                     edge_dict['Edge14-NodeSet'],
                                                                     array((vertex_dict['Vertex1-NodeSet'],
                                                                            vertex_dict['Vertex2-NodeSet'],
                                                                            vertex_dict['Vertex3-NodeSet'],
                                                                            vertex_dict['Vertex4-NodeSet']))),
                                                                    axis=None)
            simple_sets_dict['Simple-Face32-NodeSet'] = concatenate((face_dict['Face32-NodeSet'],
                                                                     edge_dict['Edge56-NodeSet'],
                                                                     edge_dict['Edge67-NodeSet'],
                                                                     edge_dict['Edge78-NodeSet'],
                                                                     edge_dict['Edge58-NodeSet'],
                                                                     array((vertex_dict['Vertex5-NodeSet'],
                                                                            vertex_dict['Vertex6-NodeSet'],
                                                                            vertex_dict['Vertex7-NodeSet'],
                                                                            vertex_dict['Vertex8-NodeSet']))),
                                                                    axis=None)
        # Define the simple node sets.
        for name, ids in simple_sets_dict.items():
            part.add_node_set(name=name, ids=ids)

    # Define the individual node sets for vertices, edges and faces.
    if explicit_sets:
        # Define node sets for the vertices.
        for name, ids in vertex_dict.items():
            part.add_node_set(name=name, ids=ids)
        # Define node sets for the edges.
        for name, ids in edge_dict.items():
            part.add_node_set(name=name, ids=ids)
        if dim.upper() == '3D':
            # Define node sets for the faces.
            for name, ids in face_dict.items():
                part.add_node_set(name=name, ids=ids)



"""Functions used for creating random Boolean masks.

The resulting masks can then be used
for manipulating a :class:`~vcams.voxelpart.VoxelPart` instance
using its :meth:`~vcams.voxelpart.VoxelPart.apply_mask` method.

See the :ref:`boolean-masks` section for an explanation of Boolean masks
and examples :doc:`C-3 </examples/example-c3>`
and :doc:`C-4 </examples/example-c4>` for sample scripts.
"""

# For some reason, the numpy.random.shuffle() function
# is added to the sphinx documentation.
# To fix this, the ":exclude-members: shuffle" directive is
# added to random-module.rst.


from logging import getLogger

from numpy import ndarray, zeros, prod, floor
from numpy.random import shuffle

logger = getLogger(__name__)


def random_binary_mask(part=None, array_shape: tuple[int, int, int] = None,
                       true_fraction: float = 0.5) -> ndarray:
    """Return a Boolean mask containing a random dispersion of True and False values.

    Args:
        part (VoxelPart | None): The *VoxelPart* instance based on which the random array is created.
                                 If *None*, *array_shape* must be specified.
        array_shape: A tuple containing three integers which determines
                    the shape of the random array. Ignored if *part* is passed.
        true_fraction: Fraction of the True values in the random array.
                       If the number of elements in the random array is not dividable,
                       it is rounded down.

    Returns:
        A Boolean mask with the desired fraction of randomly distributed True elements.
    """

    if part:
        array_shape = part.size
    num_array_elems = prod(array_shape)
    num_true = int(floor(true_fraction * num_array_elems))
    random_array = zeros(num_array_elems, dtype=bool, order='C')
    random_array[:num_true] = True
    shuffle(random_array)
    return random_array.reshape(array_shape, order='C')



__author__ = 'Mohammadreza Khoshbin'
__copyright__ = 'Copyright © 2025, Mohammadreza Khoshbin.'
__credits__ = ['Mohammadreza Khoshbin']
__license__ = 'AGPLv3'
__version__ = '3.1.1'
__maintainer__ = 'Mohammadreza Khoshbin'
__email__ = 'm.khoshbin@live.com'
__status__ = 'Production'
__website__ = 'https://github.com/mkhoshbin1/vcams'
__repo__ = 'https://github.com/mkhoshbin1/vcams'
__docs__ = 'https://vcams.readthedocs.io/'
__description__ = 'A Program and Python Library for Voxel-Based Computer-Aided Modeling of Complex Structures '
__author_website__ = 'www.mkhoshbin.com'
__contact__ = __email__
__deprecated__ = False

# The following are used in the GUI:
gui_footer_notice = (f'{__copyright__}\nVCAMS is a free and open source software published '
                     f'under the GNU AGPLv3 license.\n')
# gui_footer_notice = ('VCAMS is a free and open source software published '
#                      'under the GNU AGPLv3 license.\n')
about_vcams = ('<h1>VCAMS v%s</h1>'
               '<font size="+1"><p align="justify">VCAMS (Voxel-Based Computer-Aided '
               'Modeling of Complex Structures) is a free and open source '
               'software for creating complex FEA models using voxels. It can be used '
               'and extended by anyone in accordance with the GNU AGPLv3 license.</p>'
               '<p align="justify">You are currently using the software\'s GUI, '
               'but the main library is even more powerful! You will find links to the '
               'software\'s code and documentation (including examples) '
               'in the Help menu.</p></font>'
               '<p align="center">%s</p>' % (__version__, __copyright__))
gui_name = 'VCAMS GUI'
gui_file_name = gui_name + ' v' + __version__
gui_window_title = gui_name + ' v' + __version__
# Import the modules. Although general guidelines are against polluting the namespace,
# it's done to improve usability for less experienced users.
from . import bc  # noqa: E402
from . import mask  # noqa: E402
from . import voxelpart  # noqa: E402



"""Functions used for exporting a :class:`~vcams.voxelpart.VoxelPart` instance
for use in other programs. Currently, only Abaqus™ is supported.

These functions are not meant to be directly used.
The main function (:func:`~write_abaqus_inp`) is called
by a part's :meth:`~.voxelpart.VoxelPart.output_abaqus_inp` method,
and it uses the *VoxelPart*'s attributes for determining what is exported.

Refer to the :ref:`export` section for instructions on how to properly output a model.
"""

import logging
import os
import shutil
import time
from pathlib import Path
from typing import Union, TextIO

import numpy as np
from numpy import savetxt, unravel_index, ravel_multi_index, array, unique, uint32, float64, \
    union1d, any, zeros, append, intersect1d, insert, vstack, ndarray
# noinspection PyPackageRequirements
from tabulate import tabulate  # The main module in the tabulate2 package is named tabulate.

from . import __version__, __website__
from . import helper
from .bc import create_bc
from .helper import validate_materials_to_be_output, validate_dim

logger = logging.getLogger(__name__)


def write_abaqus_inp(part, file_name: str, elem_code: str, dim: str,
                     scale: tuple, material_elem_sets: tuple | str,
                     custom_elem_sets: bool = True, keep_temp_files: bool = False) -> Path:
    """Write a VoxelPart instance to an Abaqus™ input file.
    This is the main function called by :meth:`.voxelpart.VoxelPart.output_abaqus_inp`.
    It should not be directly used.

    Only the elements selected by the *material_elem_sets* parameter are selected,
    and afterwards they are grouped into sets by the material code.
    If *custom_elem_sets* is True, the custom element set are also included.
    If an element is part of a custom element set but is not part of the selected materials,
    It is not written to the output.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` instance on which the operation is performed.
        file_name: Name of the file. Must be valid according to the documentation
                   for the :func:`.helper.is_name_valid` function and should not contain file extensions.
        elem_code: An uppercase string denoting the element code assigned to *all* elements in the model.
                   It must be a valid Abaqus element code such as *'CPE4R'* or *'C3D8R'*.
                   This parameter is not validated so care should be taken regarding validity and compatibility.
                   Currently, only 2D and 3D linear elements are supported.
                   To get around this, you can convert to quadratic elements after importing the model to Abaqus.
        dim: Dimensionality of the output part. Valid values are *'2D'* and *'3D'*.
        scale: A tuple containing two or three floats which are used to scale
               the pixels or voxels in the x, y, and z directions.
               For example, if the tuple ``(0.02, 0.1, 1.5)`` is specified,
               each voxel will have those dimensions in the x, y, and z directions.
        material_elem_sets: One of the following:

                            + *'All'* which outputs all materials in the VoxelPart.
                            + *'Non-Empty'* which outputs all non-zero (=non-empty) materials in the VoxelPart.
                            + A tuple of integer material codes corresponding
                              to the materials that should be written to the output.

        custom_elem_sets: If set to True, custom sets will be written to the output.
        keep_temp_files: If set to True, temporary files will not be deleted. Used for debugging.

        Returns:
            Path object pointing to final Abaqus™ input file.
        """
    logger.info("Attempting to output part '%s' to an Abaqus input file.", part.name)
    begin_time = time.perf_counter()
    folder_path = part.working_dir

    # Validate dim.
    validate_dim(dim)
    if dim == '2D' and len(part.size) == 3:
        raise ValueError('Attempted to export a 3D part as a 2D part. '
                         'This is not supported due to unnecessary complexity. '
                         'Instead, create a new 2D VoxelPart instance '
                         'with the desired slice of VoxelPart.data and export it.')

    # Validate file_name and add file extension.
    if not helper.is_name_valid(file_name):
        raise ValueError('Invalid file_name. Check the documentation for validity criteria.')
    file_name = file_name + '.inp'

    # Validate and process material_elem_sets.
    material_elem_sets = validate_materials_to_be_output(part, material_list=material_elem_sets)

    # Process BCs.
    # max_empty_border_elems has not been specified.
    constraint_list = create_bc(part, dim, mat_codes_to_accept=material_elem_sets)

    # Add the dummy nodes as node sets.
    # noinspection PyProtectedMember
    for name, node_id in part._dummy_node_dict.items():
        part.add_node_set(name=name, ids=node_id - 1)  # ids is zero-based.

    # Write element sets.
    (elem_set_file_path, elem_id_list, elem_set_stats) = \
        write_elem_set_def(part, material_elem_sets, folder_path, custom_elem_sets)

    # Write temporary element definition file.
    (elem_file_path, num_elems, node_id_list) = write_elem_def(part=part,
                                                               elem_id_list=elem_id_list,
                                                               elem_code=elem_code, dim=dim,
                                                               folder_path=folder_path)

    # Write temporary node definition file.
    (node_file_path, num_nodes, node_id_list) = write_node_def(part=part,
                                                               node_id_list=node_id_list,
                                                               scale=scale, dim=dim,
                                                               folder_path=folder_path)

    # Write node sets.
    (node_set_file_path, node_set_stats) = write_node_set_def(part, node_id_list, folder_path)

    # Write constraints.
    if constraint_list:
        num_constraints, constraints_file_path = write_constraints(folder_path, constraint_list)
    else:
        num_constraints = 0

    # Write the final Abaqus input file.
    main_file_path = os.path.join(folder_path, file_name)
    logger.debug("Assembling temporary files to the main input file at '%s'.", main_file_path)
    with open(main_file_path, 'w', encoding='latin1') as main_file:
        # Write program details as a comment on top.
        main_file.write(('** Input file generated by VCAMS v%s' % __version__ +
                         time.strftime(' on %Y-%m-%d at %H:%M:%S %Z.\n', time.localtime()) +
                         '** VCAMS is a free and open source program available at:\n' +
                         ('** %s\n' % __website__) +
                         '** Author: Mohammadreza Khoshbin (www.mkhoshbin.com)\n\n'))

        # Write part description as heading text.
        main_file.write('*HEADING\n%s\n\n' % part.description)

        # Declare the parts portion of the input file as a comment.
        main_file.write('**\n** Parts\n**\n\n')

        # Declare this part.
        main_file.write('*PART, NAME="%s"\n\n' % part.name)

        # Write nodes.
        with open(node_file_path, 'r') as node_file:
            shutil.copyfileobj(node_file, main_file)

        # Write elements.
        with open(elem_file_path, 'r') as elem_file:
            shutil.copyfileobj(elem_file, main_file)

        # Write element sets.
        with open(elem_set_file_path, 'r') as elem_set_file:
            shutil.copyfileobj(elem_set_file, main_file)

        # Declare the end of this part.
        main_file.write('*END PART\n**\n\n')

        # Write an instance of the part to the assembly.
        # Declare the assembly.
        main_file.write('**\n** Assembly\n**\n\n')
        main_file.write('*ASSEMBLY, NAME=Assembly\n\n')

        # Declare the instance.
        main_file.write('*INSTANCE, NAME="%s", PART="%s"\n' % (part.instance_name, part.name))
        main_file.write('*END INSTANCE\n**\n\n')

        # Write node sets.
        with open(node_set_file_path, 'r') as node_set_file:
            shutil.copyfileobj(node_set_file, main_file)

        # Write constraints.
        if constraint_list:
            # noinspection PyUnboundLocalVariable
            with open(constraints_file_path, 'r') as constraints_file:
                shutil.copyfileobj(constraints_file, main_file)

        # Declare the end of the assembly portion of the input file.
        main_file.write('*END ASSEMBLY\n**\n\n')

    # Delete temporary file.
    if not keep_temp_files:
        logger.debug('Deleting temporary files.')
        os.remove(node_file_path)
        os.remove(elem_file_path)
        os.remove(node_set_file_path)
        os.remove(elem_set_file_path)
        if constraint_list:
            os.remove(constraints_file_path)

    elapsed_time = time.perf_counter() - begin_time
    logger.info("Finished exporting part '%s' to the Abaqus™ input file at '%s'.",
                part.name, main_file_path)

    write_output_summary(part, dim, elem_code, num_nodes, num_elems,
                         elem_set_stats, node_set_stats, num_constraints, elapsed_time)
    return Path(main_file_path)


def write_elem_def(part, elem_id_list: ndarray, elem_code: str, dim: str,
                   folder_path: str) -> tuple[str, int, ndarray]:
    """Write the element definition portion of an Abaqus™ input file to a temporary file,
    which will be concatenated with other portions to form an input file.

    Element definition consists of specifying the element code and writing its connectivity table.
    In the finite element method, the connectivity table (or matrix) determines which nodes belong to each element.
    The first column is always element id and its nodes are written
    in rest of the columns in a specific order based on the element geometry.

    Args:
        part (VoxelPart): The VoxelPart instance on which the operation is performed.
        elem_id_list: A 1-D Numpy ndarray containing IDs of elements which must be output.
                      The function makes sure that it is unique and sorted.
                      Note that Abaqus only accepts element IDs that are positive and less than 999999999.
                      Element IDs must also be integers. This is not directly checked,
                      but an error will be raised once they are passed as indices to numpy.
        elem_code: An uppercase string denoting the element code assigned to *all* elements in the model.
                   It must be a valid Abaqus element code such as 'CPE4R' or 'C3D8R'.
                   This parameter is not validated so care should be taken regarding validity and compatibility.
                   Currently, only 2D and 3D linear elements are supported.
                   To get around this, you can convert to quadratic elements after importing the model to Abaqus.
        dim: Dimensionality of the output part. Valid values are *'2D'* and *'3D'*.
        folder_path: Path to the folder where the temporary element definition file will be placed.

    Returns:
        The tuple *(file_path, num_elems, node_id_list)* containing

          #. The path to the temporary element definition file;
          #. The number of elements which have been written to the file; and
          #. A numpy ndarray containing a sorted list of node IDs that are present in the model.
    """
    logger.debug("Attempting to write element definitions to the temporary file 'elem_def.tmp'.")
    # Validate elem_id_list. Note that values are not checked.
    if len(elem_id_list) == 0:
        raise ValueError('elem_id_list is empty. At least one element must be selected for output.')
    if any(elem_id_list < 0):
        raise ValueError('At least on element in elem_id_list is negative' +
                         ' which will result in a non-positive ID in the input file.')
    if max(elem_id_list) >= 999999999:
        raise RuntimeError(('At least one element has an ID greater than 999999999,' +
                            ' which is not supported by Abaqus™.'))

    # Validate dim.
    validate_dim(dim)

    # Make sure elem_id_list is unique and sorted.
    elem_id_list = unique(elem_id_list)

    # Preallocate memory for the connectivity table.
    if dim.upper() == '2D':
        elem_array_shape = part.size[0:2]
        num_cols = 5
    elif dim.upper() == '3D':
        elem_array_shape = part.size
        num_cols = 9
    else:
        raise RuntimeError('Unexpected value for dim. This should have been caught earlier.')
    connectivity_table = zeros((elem_id_list.size, num_cols), dtype=uint32, order='C')

    # In each direction, node array is larger by one.
    node_array_shape = tuple(i + 1 for i in elem_array_shape)

    # Find the coordinates for the elements in elem_id_list.
    elem_inds = unravel_index(elem_id_list, elem_array_shape, order='C')

    # Add elem_id_list as the first column.
    connectivity_table[:, 0] = elem_id_list + 1

    # For each element in elem_id_list, find its nodes.
    if dim.upper() == '2D':
        connectivity_table[:, 1] = ravel_multi_index(multi_index=(elem_inds[0], elem_inds[1]),
                                                     dims=node_array_shape, mode='raise',
                                                     order='C') + 1
        connectivity_table[:, 2] = ravel_multi_index(multi_index=(elem_inds[0] + 1, elem_inds[1]),
                                                     dims=node_array_shape, mode='raise',
                                                     order='C') + 1
        connectivity_table[:, 3] = ravel_multi_index(
            multi_index=(elem_inds[0] + 1, elem_inds[1] + 1),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 4] = ravel_multi_index(multi_index=(elem_inds[0], elem_inds[1] + 1),
                                                     dims=node_array_shape, mode='raise',
                                                     order='C') + 1
    elif dim.upper() == '3D':
        connectivity_table[:, 1] = ravel_multi_index(
            multi_index=(elem_inds[0], elem_inds[1], elem_inds[2]),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 2] = ravel_multi_index(
            multi_index=(elem_inds[0] + 1, elem_inds[1], elem_inds[2]),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 3] = ravel_multi_index(
            multi_index=(elem_inds[0] + 1, elem_inds[1] + 1, elem_inds[2]),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 4] = ravel_multi_index(
            multi_index=(elem_inds[0], elem_inds[1] + 1, elem_inds[2]),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 5] = ravel_multi_index(
            multi_index=(elem_inds[0], elem_inds[1], elem_inds[2] + 1),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 6] = ravel_multi_index(
            multi_index=(elem_inds[0] + 1, elem_inds[1], elem_inds[2] + 1),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 7] = ravel_multi_index(
            multi_index=(elem_inds[0] + 1, elem_inds[1] + 1, elem_inds[2] + 1),
            dims=node_array_shape, mode='raise', order='C') + 1
        connectivity_table[:, 8] = ravel_multi_index(
            multi_index=(elem_inds[0], elem_inds[1] + 1, elem_inds[2] + 1),
            dims=node_array_shape, mode='raise', order='C') + 1
    else:
        raise RuntimeError('Unexpected value for dim. This should have been caught earlier.')

    # Write the element connectivity table to a temporary text file.
    file_path = os.path.join(folder_path, 'elem_def.tmp')
    # noinspection PyTypeChecker
    savetxt(fname=file_path, X=connectivity_table,
            fmt='%u', delimiter=',', comments='', encoding='latin1',
            header=('*ELEMENT, TYPE=%s' % elem_code.upper()), footer='\n')

    # Create a sorted array of unique node IDs in the connectivity matrix
    # and revert node IDs to zero-based indexing.
    node_id_list = unique(connectivity_table[:, 1:]) - 1
    num_elems = len(elem_id_list)

    logger.debug(f"Wrote {num_elems} {dim.upper()} elements "
                 f"of type '{elem_code.upper()}' to the temporary file 'elem_def.tmp'.")
    return file_path, num_elems, node_id_list


def write_node_def(part, node_id_list: ndarray, scale: tuple, dim: str, folder_path: str) -> tuple[str, int, ndarray]:
    """Write the node definition portion of an Abaqus™ input file to a temporary file,
    which will be concatenated with other portions to form an input file.

    Node definition consists of specifying the node ID and writing its coordinates in the x, y, and z directions.
    The code uses a global cartesian coordinate system which is compatible with its data structure and sufficient
    for our purposes.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` instance on which the operation is performed.
        node_id_list: A 1-D Numpy ndarray containing IDs of nodes which must be output.
                      The function makes sure that it is unique and sorted.
                      Note that Abaqus only accepts node IDs that are positive and less than 999999999
                      and this is reduced to 999999990 to account for possible dummy nodes.
                      Node IDs must also be integers. This is not directly checked,
                      but an error will be raised once they are passed as indices to numpy.
        scale: A tuple containing two or three floats which are used to scale
               the pixels or voxels in the x, y, and z directions.
               For example, if the tuple ``(0.02, 0.1, 1.5)`` is specified,
               each voxel will have those dimensions in the x, y, and z directions.
        dim: Dimensionality of the output part. Valid values are *'2D'* and *'3D'*.
        folder_path: Path to the folder where the temporary node definition file will be placed.

    Returns:
        The tuple *(file_path, num_nodes, node_id_list)* containing

          #. The path to the temporary node definition file;
          #. The number of nodes which have been written to the file; and
          #. A numpy ndarray containing the IDs of the nodes written to file
             which has been updated by adding the dummy nodes
    """
    logger.debug("Attempting to write node definitions to the temporary file 'node_def.tmp'.")
    # Validate node_id_list.
    if len(node_id_list) == 0:
        raise ValueError('node_id_list is empty. At least one node must be selected for output.')
    if any(node_id_list < 0):
        raise ValueError('At least on element in node_id_list is negative' +
                         ' which will result in a non-positive ID in the input file.')
    if max(node_id_list) >= 999999990:
        raise RuntimeError(('At least one node has an ID greater than 999999990,' +
                            ' which is not supported by Abaqus™.'))
    num_real_nodes = node_id_list.size

    # Validate dim.
    validate_dim(dim)

    # Get dummy_node_dict from the part.
    # noinspection PyProtectedMember
    dummy_node_dict = part._dummy_node_dict

    # Set format string and number of columns in node_coordinates based on dim.
    if dim.upper() == '2D':
        num_cols = 3
        fmt = [('%' + str(len(str(max(node_id_list)))) + 'u'), '%13.12G', '%13.12G']
    elif dim.upper() == '3D':
        num_cols = 4
        fmt = [('%' + str(len(str(max(node_id_list)))) + 'u'), '%13.12G', '%13.12G', '%13.12G']
    else:
        raise RuntimeError('Unexpected value for dim. This should have been caught earlier.')

    # Preallocate memory for node_coordinates.
    # The dummy nodes are added at the end, so allocate accordingly.
    node_table = zeros((num_real_nodes + len(dummy_node_dict), num_cols), dtype=float64, order='C')

    # Add node_id_list as the first column of node_table.
    node_table[:num_real_nodes, 0] = node_id_list + 1

    # Obtain Cartesian coordinates of each node by unraveling its index and multiplying it by scale.
    node_array_shape = tuple(i + 1 for i in part.size)
    raw_indices = unravel_index(node_id_list, node_array_shape, order='C')

    # Add node coordinates to node_table.
    node_table[:num_real_nodes, 1] = raw_indices[0] * scale[0]
    node_table[:num_real_nodes, 2] = raw_indices[1] * scale[1]
    if dim.upper() == '3D':
        node_table[:num_real_nodes, 3] = raw_indices[2] * scale[2]

    if dummy_node_dict:
        # Add the dummy node to the end of the table.
        dummy_nodes_list = []
        max_size = node_table.max(axis=0, initial=-1)
        offset = 0.05
        for name, node_id in dummy_node_dict.items():
            if name == 'RP0-NodeSet':
                dummy_nodes_list.append((insert((max_size * -0.05)[1:], 0, node_id, axis=None)))
            else:
                dummy_nodes_list.append((insert((max_size * (1 + offset))[1:], 0, node_id, axis=None)))
                offset += 0.05
        dummy_nodes_list_array = vstack(dummy_nodes_list)
        node_table[-1 * len(dummy_node_dict):, :] = dummy_nodes_list_array[dummy_nodes_list_array[:, 0].argsort()]
        node_id_list = append(node_id_list, [i - 1 for i in dummy_node_dict.values()])

    # Write node_table to a temporary text file.
    file_path = os.path.join(folder_path, 'node_def.tmp')
    # noinspection PyTypeChecker
    savetxt(fname=file_path, X=node_table,
            fmt=fmt, delimiter=',', comments='', encoding='latin1',
            header='*NODE', footer='\n')

    logger.debug("Wrote %u %s nodes to the temporary file 'node_def.tmp'.",
                 num_real_nodes, dim.upper())
    return file_path, num_real_nodes, node_id_list


def write_set_ids(file_obj: TextIO, kind: str, name: str, ids: ndarray,
                  instance_name: Union[str, None] = None) -> tuple[str, int]:
    """Write an element or node set to a file according to Abaqus™ input file syntax.

    Args:
        file_obj: The file object in which the set is written. It must be opened in text mode.
        kind: The kind of set that is to be output.
              Valid values are *'ELSET'* for an element set and *'NSET'* for a node set.
        name: Name of the set. Must be valid according to
              the documentation for the :func:`.helper.is_name_valid` function.
        ids: A 1-D Numpy ndarray containing zero-based IDs of nodes or elements which form the set.
             The function makes sure that it is unique and sorted.
             Note that Abaqus only accepts IDs that are positive and less than 999999999.
             IDs must also be integers, but this is not validated in this function.
        instance_name: If set to a string, the set is defined with an *'INS'* parameter
                       to signify it belonging to an instance.
                       If set to *None*, the set is defined without it which signifies it belonging to a part.

    Returns:
        A tuple containing set name and number of IDs in the set.
    """
    if kind.upper() not in ['ELSET', 'NSET']:
        raise ValueError("kind can only be one of 'ELSET' or 'NSET'.")

    if not helper.is_name_valid(name):
        raise ValueError('Invalid name. Check the documentation for validity criteria.')

    # Validate ids. Note that values are not checked.
    if len(ids) == 0:
        raise ValueError('ids is empty. At least one ID must be selected for output.')
    if any(ids < 0):
        raise ValueError('At least on element in ids is negative '
                         'which will result in a non-positive ID in the input file.')
    if max(ids) >= 999999999:
        raise RuntimeError('At least one ID is greater than 999999999, '
                           'which is not supported by Abaqus™.')
    # Make sure ids is unique and sorted.
    ids = unique(ids)

    # IDs start at zero (zero-based indexing).
    # Add the number 1 to all of them to account for that.
    ids = ids + 1

    # Write the set header manually.
    if instance_name:
        file_obj.write('*%s,%s="%s",INS="%s"\n' % (kind.upper(), kind.upper(), name, instance_name))
    else:
        file_obj.write('*%s,%s="%s"\n' % (kind.upper(), kind.upper(), name))

    # If there are 9 or fewer IDs, they are written manually.
    # Otherwise, they are written as chunks of 9 IDs, ensuring low line length.
    # numpy.savetxt can only write a 2D array, meaning that the extra IDs
    # need to be written separately.
    # For simplicity, the last 9 elements are always written separately.
    if len(ids) <= 9:
        extra_chunk = ids
    else:
        num_extra = ids.size % 9
        if num_extra == 0:
            num_extra = 9
        main_chunk = ids[:-num_extra].reshape((-1, 9))
        extra_chunk = ids[-num_extra:]
        # Write the main chunk using numpy.
        savetxt(fname=file_obj, X=main_chunk,
                fmt='%u', delimiter=',', newline=',\n', comments='', encoding='latin1',
                header='', footer='')

    # Write the extra chunk manually.
    file_obj.write(','.join(['%u'] * len(extra_chunk)) % tuple(extra_chunk))
    file_obj.write('\n\n')

    return name, len(ids)


def write_elem_set_def(part, material_elem_sets: tuple, folder_path: str,
                       custom_elem_sets: bool = True) -> tuple[str, ndarray, dict]:
    """Write the element set portion of an Abaqus™ input file to a temporary file.
    This function also returns which elements must be output.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` instance on which the operation is performed.
        material_elem_sets: A tuple containing integer values of the materials that should be output.
                            For each material *x*, a set named *'MAT-x'* is defined.
        custom_elem_sets: If set to True, the custom sets are defined and output.
        folder_path: Path to the folder where the temporary elemset definition file will be placed.

    Returns:
        The tuple *(elem_set_file_path, elem_id_list, elem_set_stats)* containing

          #. The path to the temporary element set definition file;
          #. A ndarray containing a unique and sorted list of all element IDs in the sets
          #. A dictionary where the keys are names of the element sets
             and the values are the number of elements in that set.
    """
    logger.debug("Attempting to write element sets to the temporary file 'elemset.tmp'.")
    elem_set_stats = dict()
    elem_id_list = array([], order='C', dtype='uint32')
    elem_set_file_path = os.path.join(folder_path, 'elemset.tmp')

    with open(elem_set_file_path, 'w', encoding='latin1') as file_obj:
        # Write custom element sets. The element IDs are unique and sorted.
        if custom_elem_sets and bool(part.elem_sets):
            for (name, elem_ids) in part.elem_sets.items():
                if elem_ids.size > 0:
                    (set_name, num_ids) = write_set_ids(file_obj=file_obj, kind='ELSET',
                                                        name=name, ids=elem_ids)
                    elem_id_list = union1d(elem_id_list, elem_ids)
                    elem_set_stats[set_name] = num_ids
        # Write the materials that should be output.
        for mat_code in material_elem_sets:
            # noinspection PyProtectedMember
            (name, elem_ids) = part._return_material_elem_set(mat_code)
            if elem_ids.size > 0:
                (set_name, num_ids) = write_set_ids(file_obj=file_obj, kind='ELSET',
                                                    name=name, ids=elem_ids)
                elem_id_list = union1d(elem_id_list, elem_ids)
                elem_set_stats[set_name] = num_ids

    logger.debug("Wrote %u element sets to the temporary file 'elemset.tmp'.", len(elem_set_stats))
    return elem_set_file_path, elem_id_list, elem_set_stats


def write_node_set_def(part, node_id_list: ndarray, folder_path: str) -> tuple[str, dict]:
    """Write the node set portion of an Abaqus™ input file to a temporary file.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` instance on which the operation is performed.
        node_id_list: A 1-D Numpy ndarray containing IDs of the nodes that have been written to the output.
                      It is used to determine which nodes from the node sets should be defined in the output.
        folder_path: Path to the folder where the temporary nodeset definition file will be placed.

    Returns:
        The path to the temporary node set definition file.
    """
    logger.debug("Attempting to write node sets to the temporary file 'nodeset.tmp'.")
    node_set_stats = dict()
    node_set_file_path = os.path.join(folder_path, 'nodeset.tmp')
    with open(node_set_file_path, 'w', encoding='latin1') as file_obj:
        # Write node sets.
        num_omitted = 0
        for (name, node_ids) in part.node_sets.items():
            node_set_ids = intersect1d(node_id_list, node_ids)
            if len(node_set_ids) == 0:
                num_omitted += 1
                logger.warning("Node set '%s' was not written to output because none of "
                               "its nodes and elements are set for output.", name)
            else:
                # noinspection PyTypeChecker
                write_set_ids(file_obj=file_obj, kind='NSET', name=name,
                              ids=node_set_ids, instance_name=part.instance_name)
                node_set_stats[name] = len(node_set_ids)

    logger.debug("Wrote %u node sets to the temporary file 'nodeset.tmp'.",
                 len(part.node_sets) - num_omitted)
    return node_set_file_path, node_set_stats


def write_constraints(folder_path: str, constraint_list: list) -> tuple[int, str]:
    """Write the defined constraints portion of an Abaqus™ input file to a temporary file.

    Args:
        constraint_list: Tuple of constraint objects defined in :doc:`bc-module`.
                         Their `repr()` function is written to the file.
        folder_path: Path to the folder where the temporary constraint definition file will be placed.

    Returns:
        A tuple containing the number of constraints and the path to the temporary constraint definition file.
    """
    # Note: Creation of a 2D PBC with 16M elements was tested.
    # Since the number of constrain equations are actually small (16000),
    # This function takes almost no time and so no further optimization is necessary.
    logger.debug("Attempting to write constraints to the temporary file 'constraints_def.tmp'.")
    constraints_file_path = os.path.join(folder_path, 'constraints_def.tmp')
    with open(constraints_file_path, 'w', encoding='latin1') as file_obj:
        file_obj.write('**\n** Constraints\n')
        for constraint_obj in constraint_list:
            file_obj.write(repr(constraint_obj))
        file_obj.write('** End Constraints\n\n')
    num_constraints = repr(constraint_list[0]).count('*Equation') * len(constraint_list)
    logger.debug("Wrote %u constraints to the temporary file 'constraints_def.tmp'.", num_constraints)
    return num_constraints, constraints_file_path


def write_output_summary(part, dim: str, elem_code: str, num_nodes: int, num_elems: int,
                         elem_set_stats: dict, node_set_stats: dict, num_constraints: int,
                         elapsed_time: float):
    """Write a summary of the output to the main log.

    Args:
        part (VoxelPart): The :class:`~.voxelpart.VoxelPart` instance on which the operation is performed.
        dim: Dimensionality of the output part. Valid values are *'2D'* and *'3D'*.
        elem_code: An uppercase string denoting the element code assigned to *all* elements in the model.
                   See :func:`write_abaqus_inp` for complete description.
        num_nodes: Number of nodes written to the output.
        num_elems: Number of elements written to the output.
        elem_set_stats: The dict object returned by :func:`write_elem_set_def`.
        node_set_stats: The dict object returned by :func:`write_node_set_def`.
        num_constraints: Number of constraint equations written to the output.
        elapsed_time: Elapsed time for the output process in seconds.
    """

    # noinspection PyProtectedMember
    part_bc_type = part._bc_type
    if part_bc_type is None:
        bc_name = 'None'
    elif part_bc_type.upper() == 'NODESET ONLY':
        bc_name = 'Nodeset Only'
    elif part_bc_type.upper() == 'LINEAR DISPLACEMENT':
        bc_name = 'Linear Displacement'
    elif part_bc_type.upper() == 'PERIODIC':
        bc_name = 'Periodic Boundary Conditions'
    else:
        raise RuntimeError(f'part has an invalid bc_type {part_bc_type}. '
                           f'This should have been caught earlier.')

    # Prepare part summary.
    part_summary = (
        ('Part Name', part.name),
        ('Number of Voxels', ' * '.join(str(i) for i in part.size)),
        ('Voxel Size', ' * '.join(str(i) for i in part.voxel_size)),
        ('Part Dimensions', ' * '.join(str(i) for i in part.real_size)),
        ('Output Dimensionality', dim.upper()),
        ('Element Type', elem_code),
        ('Number of Elements', num_elems),
        ('Number of Nodes', num_nodes),
        ('Number of Element Sets', len(elem_set_stats)),
        ('Number of Node Sets', len(node_set_stats)),
        ('Number of Material Codes', len(np.unique(part.data))),
        ('Requested BC Constraints', bc_name),
        ('Number of Constraint Equations', num_constraints),
        ('Total Output Time', time.strftime('%H:%M:%S', time.gmtime(elapsed_time)))
    )

    # Prepare set stats.
    mat_elem_sets = []
    custom_elem_sets = []
    elem_set_stats_list = sorted(elem_set_stats.items(), key=lambda x: x[1], reverse=True)
    node_set_stats_list = sorted(node_set_stats.items(), key=lambda x: x[1], reverse=True)
    for item in elem_set_stats_list:
        if item[0].upper().startswith('MAT-'):
            mat_elem_sets.append((item[0], item[1],
                                  '{:.2f}'.format((item[1] / num_elems) * 100),
                                  '{:.2f}'.format((item[1] / part.data.size) * 100)))
        else:
            custom_elem_sets.append(
                (item[0], item[1], '{:.2f}'.format((item[1] / num_elems) * 100)))
    summary_text = (
            '*Model Details*\n' +
            (tabulate(part_summary, tablefmt='pretty', colalign=('left', 'left')) + '\n') +
            '\n*Material Element Sets*\n' +
            (tabulate(mat_elem_sets, headers=('Set Name', 'Number of Elements',
                                              'Percent of All Elements', 'Percent of Model'),
                      tablefmt='pretty', colalign=('left', 'left', 'left')) + '\n')
    )
    if len(custom_elem_sets) == 0:
        summary_text += '\n*Custom Element Sets*\nNo custom element sets were defined.\n'
    else:
        summary_text += ('\n*Custom Element Sets*\n' +
                         tabulate(custom_elem_sets,
                                  headers=('Set Name', 'Number of Elements',
                                           'Percent of All Elements'),
                                  tablefmt='pretty') + '\n'
                         )
    if len(node_set_stats_list) == 0:
        summary_text += '\n*Node Sets*\nNo node sets were defined.\n'
    else:
        summary_text += ('\n*Node Sets*\n' +
                         tabulate(node_set_stats_list,
                                  headers=('Set Name', 'Number of Nodes'),
                                  tablefmt='pretty') + '\n'
                         )
    logger.info('A summary of the created part is as follows:\n***\n%s***\n' % summary_text)
